// $Id: event.cpp,v 1.26 2007/04/09 20:33:46 erynes Exp $

/* 
Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include <algorithm>
#include "branch.h"
#include "shared_ptr.hpp"
#include "event.h"
#include "forceparam.h"
#include "forcesummary.h"
#include "arranger.h"
#include "tree.h"
#include "registry.h"
#include "mathx.h"
#include "stringx.h"
#include "timesize.h"
#include <functional> // for divides<> and less<> used in ThrowIfSubPopSizeTiny

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

using namespace std;

typedef boost::shared_ptr<RBranch> RBranch_ptr;


//____________________________________________________________________________
//____________________________________________________________________________

Event::Event(const ForceSummary& fs)
	: ar(NULL)
{
   m_CalcSizeAtTime = fs.GetSizeCalculator();
} /* Event ctor */

//____________________________________________________________________________

bool Event::Done() const
{ 
// As far as this event knows, we are Done if there are no more active 
// lineages.

  if (ar->ActiveSize() == 0) return true;
  else return false;

} /* Done */

//____________________________________________________________________________

Event::Event(const Event& src) :
ar(src.ar), m_maxEvents(src.m_maxEvents), m_pindex(src.m_pindex),
m_thetas(src.m_thetas), m_growths(src.m_growths)
{
  if (src.m_CalcSizeAtTime) {
     m_CalcSizeAtTime = src.m_CalcSizeAtTime->Clone();
  } else {
     // Clone (below) shouldn't handle NULL pointers, but NULL
     // is a perfectly valid value for this field and we want
     // to preserve it rather than introducing garbage into memory.
     m_CalcSizeAtTime = NULL;
  }
} /* Event copy ctor */

//____________________________________________________________________________

void Event::InstallParameters(const ForceParameters& starts)
{
  m_thetas = starts.GetRegionalThetas();
  m_growths = starts.GetGrowthRates();
} /* Event::InstallParameters */

//____________________________________________________________________________

void Event::InstallThetas(const DoubleVec1d& thetas)
{
  m_thetas = thetas;
} /* InstallThetas */

//____________________________________________________________________________
//____________________________________________________________________________

XPartitionEvent::XPartitionEvent(const ForceSummary& fs)
	: Event(fs),
          m_nxparts(registry.GetDataPack().GetNCrossPartitions()) 
{
       	/* deliberately blank */
} /* XPartitionEvent ctor */

//____________________________________________________________________________

void XPartitionEvent::ThrowIfSubPopSizeTiny(double eventT) const
{
  DoubleVec1d popsizes(m_CalcSizeAtTime->
                       SizeAt(m_thetas,m_growths,eventT));
  transform(popsizes.begin(),popsizes.end(),popsizes.begin(),
            bind2nd(divides<double>(),defaults::minMuRate));
  if (std::find_if(popsizes.begin(),popsizes.end(),bind2nd(less<double>(),1.0))
      != popsizes.end()) {
     std::string estring("XPartitionEvent::ThrowIfSubPopSizeTiny:  ");
     estring += "Tried to have a coal event at " + ToString(eventT);
     tinypopulation_error e(estring);
     throw e;
  }

} /* XPartitionEvent::ThrowIfSubPopSizeTiny */

//____________________________________________________________________________
//____________________________________________________________________________

CoalEvent::CoalEvent(const ForceSummary& fs, bool isactive) 
  : XPartitionEvent(fs),
    m_chosenxpart(FLAGLONG),
    m_isactive(isactive)
{
	// set up base class fields deferred to the subclass
    m_maxEvents = fs.GetMaxEvents(force_COAL);
    m_pindex = FLAGLONG; // we use cross partitions, not partitions
} /* CoalEvent constructor */

//____________________________________________________________________________

void CoalEvent::InstallParameters(const ForceParameters& starts)
{
  Event::InstallParameters(starts);
  
} /* InstallParameters */

//____________________________________________________________________________

void CoalEvent::DoEvent(double eventT)
{
  ThrowIfSubPopSizeTiny(eventT); 
  // Pick two active branches from the population at random.
  double rn = ar->randomSource->Float();
  Branch_ptr branch1 = ar->m_activelist.RemoveBranch(force_COAL, m_chosenxpart, rn);

  rn = ar->randomSource->Float();
  Branch_ptr branch2;

  if (m_isactive) {
    branch2 = ar->m_activelist.RemoveBranch(force_COAL, m_chosenxpart, rn);
  }
  else {
    branch2 = ar->m_inactivelist.RemoveBranch(force_COAL, m_chosenxpart, rn);
  }
  FinishEvent(eventT, branch1, branch2);
} /* DoEvent */

void CoalEvent::FinishEvent(double eventT, Branch_ptr branch1, Branch_ptr branch2)
{
  assert(branch1->HasSamePartitionsAs(branch2));
  if (m_isactive) {
    Branch_ptr newbranch = ar->GetTree()->
      CoalesceActive(eventT, m_maxEvents, branch1, branch2);
    ar->m_activelist.Append(newbranch);
  }
  else {
    Branch_ptr newbranch = ar->GetTree()->
      CoalesceInactive(eventT, m_maxEvents, branch1, branch2);
    ar->m_inactivelist.Collate(newbranch);
  }
} /* FinishEvent */

//____________________________________________________________________________
//____________________________________________________________________________

ActiveCoal::ActiveCoal(const ForceSummary& fs) 
  : CoalEvent(fs, true),
    m_invTheta()
{
} /* ActiveCoal constructor */

//____________________________________________________________________________

Event* ActiveCoal::Clone() const
{
  ActiveCoal* event = new ActiveCoal(*this);
  return event;

} /* ActiveCoal::Clone */

//____________________________________________________________________________

void ActiveCoal::InstallParameters(const ForceParameters& starts)
{
  CoalEvent::InstallParameters(starts);

  m_invTheta = starts.GetRegionalThetas();
  
  // store 1/Theta
  std::transform(m_invTheta.begin(), m_invTheta.end(),
		  m_invTheta.begin(),
		  std::bind1st(std::divides<double>(),1.0));

} /* InstallParameters */

//____________________________________________________________________________

void ActiveCoal::InstallThetas(const DoubleVec1d& thetas)
{
  Event::InstallThetas(thetas);
  m_invTheta = thetas;
  
  // store 1/Theta
  std::transform(m_invTheta.begin(), m_invTheta.end(),
		  m_invTheta.begin(),
		  std::bind1st(std::divides<double>(),1.0));

} /* InstallThetas */

//____________________________________________________________________________

double ActiveCoal::PickTime(double)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of this "delta t" is theta/(k*(k-1)),
  // where k (the number of active lineages) and theta correspond to the
  // population which yields the smallest "delta t."

  map<double, long> times;
  double newtime;
  long i;

  for (i = 0; i < m_nxparts; ++i) // loop over populations
    {
      // if there are at least two active lineages in this population...
      if (ar->m_xactives[i] > 1)
	{
	  // compute the time of the next coalescence
	  newtime = - log(ar->randomSource->Float()) / 
	    (m_invTheta[i] * ar->m_xactives[i] * (ar->m_xactives[i] - 1.0));

	  // insert it into map (for sorting)
	  times.insert(make_pair(newtime, i));
	}
    }

  if (times.empty())
    {
      // no event is possible
      m_chosenxpart = FLAGLONG;
      return FLAGDOUBLE;
    }
  else
    {
      // the first map entry is the smallest, and thus chosen, time
      map<double, long>::const_iterator mapit = times.begin();
      m_chosenxpart = (*mapit).second;
      assert ((*mapit).first >= 0.0);
      return (*mapit).first;
    }

} /* ActiveCoal::PickTime */

//____________________________________________________________________________
//____________________________________________________________________________

ActiveGrowCoal::ActiveGrowCoal(const ForceSummary& fs) 
  : CoalEvent(fs, true)
{
} /* ActiveGrowCoal constructor */

//____________________________________________________________________________

Event* ActiveGrowCoal::Clone() const
{
  ActiveGrowCoal* event = new ActiveGrowCoal(*this);
  return event;

} /* ActiveCoal::Clone */

//____________________________________________________________________________

double ActiveGrowCoal::PickTime(double starttime)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of this "delta t" is:
  //                           (1/g)*ExpE1(k*(k-1)*exp(g*starttime)/(g*theta)),
  // where g, theta, and k are the growth rate, theta, and number of active lineages
  // for the population which yields the smallest "delta t."
  // For details on the function ExpE1(), see mathx.cpp.

  map<double, long> times;
  double newtime = 0.0;
  double magicstart;
  long i;

  // compute coalescence terms per cross partition
  for (i = 0; i < m_nxparts; ++i)
    {
      // if there are at least two active lineages in this population...
      if (ar->m_xactives[i] > 1)
	{
	  // compute the time of the next coalescence
	  // first do computation in "magic time"
	  newtime =  - log(ar->randomSource->Float()) / 
	    (ar->m_xactives[i] * (ar->m_xactives[i] - 1.0) / m_thetas[i]);

	  // now convert to "real time"
	  if (m_growths[i] != 0)
	    {
	      // convert starttime to magical time
	      magicstart = -(1.0 - exp(m_growths[i] * starttime)) / m_growths[i];
	      // add starttime to displacement
	      newtime += magicstart;
	      // comvert result to real time
	      if (1.0 + m_growths[i] * newtime > 0.0)
		{ 
		  newtime = log(1.0 + m_growths[i] * newtime)/m_growths[i];
		  // convert back to displacement
		  newtime -= starttime;
		  // guard against underflow
		  if (newtime < DBL_EPSILON)
		    {
		      tinypopulation_error e("bad time proposed in Event::PickTime");
		      throw e;
		    }
		}
	      else
		newtime = MAX_LENGTH; // make this interval length
	                              // very unlikely to win the race
	    }
	  // (else newtime holds the same number that would have been
	  // computed by ActiveCoal, in the absence of growth)

	  // insert it into map (for sorting)
	  times.insert(make_pair(newtime, i));
	}
    }

  if (times.empty())
    {
      // no event is possible
      m_chosenxpart = FLAGLONG;
      return FLAGDOUBLE;
    }
  else
    {
      // the first map entry is the smallest, and thus chosen, time
      map<double, long>::const_iterator mapit = times.begin();
      m_chosenxpart = (*mapit).second;
      assert ((*mapit).first >= 0.0);
      return (*mapit).first;
    }

} /* ActiveGrowCoal::PickTime */


//____________________________________________________________________________
//____________________________________________________________________________

ActiveLogisticSelectionCoal::ActiveLogisticSelectionCoal(const ForceSummary& fs) 
  : CoalEvent(fs, true), m_selectionCoefficient(0.)
{
} /* ActiveLogisticSelectionCoal constructor */

//____________________________________________________________________________

Event* ActiveLogisticSelectionCoal::Clone() const
{
  ActiveLogisticSelectionCoal* event = new ActiveLogisticSelectionCoal(*this);
  return event;

} /* ActiveLogisticSelectionCoal::Clone */


//____________________________________________________________________________

void ActiveLogisticSelectionCoal::InstallParameters(const ForceParameters& starts)
{
  CoalEvent::InstallParameters(starts);
  m_selectionCoefficient = starts.GetLogisticSelectionCoefficient()[0];
  LogisticSelectionTimeSize *pLSTS = dynamic_cast<LogisticSelectionTimeSize*>(Event::m_CalcSizeAtTime);
  if (!pLSTS)
    {
      string msg = "ActiveLogisticSelectionCoal::InstallParameters(), unable ";
      msg += "to dynamically cast its TimeSize pointer to a LogisticSelectionTimeSize ";
      msg += "pointer.";
      throw implementation_error(msg);
    }
  pLSTS->SetSelectionCoefficient(m_selectionCoefficient);
} /* InstallParameters */

//____________________________________________________________________________
//____________________________________________________________________________

ActiveStairStickCoal::ActiveStairStickCoal(const ForceSummary& fs) 
  : CoalEvent(fs, true)
{
} /* ActiveStairStickCoal constructor */

//____________________________________________________________________________

Event* ActiveStairStickCoal::Clone() const
{
  ActiveStairStickCoal* event = new ActiveStairStickCoal(*this);
  return event;

} /* ActiveStairStickCoal::Clone */

//____________________________________________________________________________

double ActiveStairStickCoal::PickTime(double starttime)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time
  // interval, not the timestamp for the end of the interval.

  TimeList& intervals(ar->m_tree->GetTimeList());
  JBranch_ptr joint(intervals.FindStickAtTime(starttime));
  DoubleVec1d newfrac(m_nxparts);
  DoubleVec1d::iterator it;
  for(it = newfrac.begin(); it != newfrac.end(); ++it)
     // technically, we should be taking the log of (1-rnd[float]),
     // but that's the same as log(rnd[float]), so omitted for
     // simplicity/speed
     (*it) = -1.0 * log(ar->randomSource->Float());

  if (*max_element(ar->m_xactives.begin(),ar->m_xactives.end()) < 2) {
     m_chosenxpart = FLAGLONG;
     return FLAGDOUBLE;
  }

  DoubleVec1d total(newfrac.size(),0.0), oldtotal(newfrac.size(),0.0);
  if (!joint->NextJoint()) intervals.AddNextJoint(joint);
  double length(joint->BottomTime() - starttime), totallength(0.0);

  while(true) {
    oldtotal = total;
    DoubleVec1d thetas(joint->GetThetas());

    assert(thetas.size() == newfrac.size());

    double mindisplace(DBL_MAX);
    long i;
    for (i = 0; i < m_nxparts; ++i) {
       if (ar->m_xactives[i] < 2) continue;
       total[i] += ar->m_xactives[i] * (ar->m_xactives[i] - 1.0) /
          thetas[i] * length;
       if (total[i] > newfrac[i]) {
          double newdisplace(newfrac[i] - oldtotal[i]);
	  newdisplace /= ar->m_xactives[i] * (ar->m_xactives[i] - 1.0) /
             thetas[i];
	  newdisplace += totallength;
	  if (newdisplace < mindisplace) {
             m_chosenxpart = i;
	     mindisplace = newdisplace;
	  }
       }
    }
    if (mindisplace < DBL_MAX) return mindisplace;
    totallength += length;
    if (totallength >= MAX_LENGTH) return MAX_LENGTH;

    joint = intervals.NextJointAppendingIfLast(joint);
    if (!joint->NextJoint()) intervals.AddNextJoint(joint);
    length = joint->BottomTime() - joint->m_eventtime;
  }

  // can't get here
  assert(false);
  m_chosenxpart = FLAGLONG;
  return FLAGDOUBLE;
 
} /* ActiveStairStickCoal::PickTime */

//____________________________________________________________________________
//____________________________________________________________________________

InactiveStairStickCoal::InactiveStairStickCoal(const ForceSummary& fs) 
  : CoalEvent(fs, false)
{
} /* InactiveStairStickCoal constructor */

//____________________________________________________________________________

Event* InactiveStairStickCoal::Clone() const
{
  InactiveStairStickCoal* event = new InactiveStairStickCoal(*this);
  return event;

} /* InactiveStairStickCoal::Clone */

//____________________________________________________________________________

double InactiveStairStickCoal::PickTime(double starttime)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time
  // interval, not the timestamp for the end of the interval.

  TimeList& intervals(ar->m_tree->GetTimeList());
  JBranch_ptr joint(intervals.FindStickAtTime(starttime));
  DoubleVec1d newfrac(ar->m_xactives.size());
  DoubleVec1d::iterator it;
  for(it = newfrac.begin(); it != newfrac.end(); ++it)
     (*it) = -1.0 * log(ar->randomSource->Float());

  if (*max_element(ar->m_xactives.begin(),ar->m_xactives.end()) < 1 ||
      *max_element(ar->m_xinactives.begin(),ar->m_xinactives.end()) < 1 ) {
     m_chosenxpart = FLAGLONG;
     return FLAGDOUBLE;
  }

  DoubleVec1d total(newfrac.size(),0.0), oldtotal(newfrac.size(),0.0);
  if (!joint->NextJoint()) intervals.AddNextJoint(joint);
  double length(joint->BottomTime() - starttime), totallength(0.0);

  while(true) {
    oldtotal = total;
    DoubleVec1d thetas(joint->GetThetas());

    assert(thetas.size() == newfrac.size());

    double mindisplace(DBL_MAX);
    long i;
    for (i = 0; i < m_nxparts; ++i) {
       if (ar->m_xinactives[i] < 1 || ar->m_xactives[i] < 1) continue;
       total[i] += 2.0 * ar->m_xactives[i] * ar->m_xinactives[i] /
          thetas[i] * length;
       if (total[i] > newfrac[i]) {
          double newdisplace(newfrac[i] - oldtotal[i]);
	  newdisplace /= 2.0 * ar->m_xactives[i] * ar->m_xinactives[i] /
             thetas[i];
	  newdisplace += totallength;
	  if (newdisplace < mindisplace) {
             m_chosenxpart = i;
	     mindisplace = newdisplace;
	  }
       }
    }
    if (mindisplace < DBL_MAX) return mindisplace;
    totallength += length;
    if (totallength >= MAX_LENGTH) return MAX_LENGTH;

    joint = intervals.NextJointAppendingIfLast(joint);
    if (!joint->NextJoint()) intervals.AddNextJoint(joint);
    length = joint->BottomTime() - joint->m_eventtime;
  }

  // can't get here
  assert(false);
  m_chosenxpart = FLAGLONG;
  return FLAGDOUBLE;
 
} /* InactiveStairStickCoal::PickTime */


double ActiveLogisticSelectionCoal::PickTime(double starttime)
{
  if (2 != m_nxparts)
    {
      string msg = "ActiveLogisticSelectionCoal::PickTime() called with m_nxparts = ";
      msg += ToString(m_nxparts);
      msg += ".  m_nxparts must equal 2, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  if (ar->m_xactives[0] <= 1 && ar->m_xactives[1] <= 1)
    {
      // no event is possible
      m_chosenxpart = FLAGLONG;
      return FLAGDOUBLE;
    }

  double theta_A0(m_thetas[0]), theta_a0(m_thetas[1]), s(m_selectionCoefficient);
  double dtau_A, dtau_a, dtA_1(FLAGDOUBLE), dtA_2(FLAGDOUBLE), dta_1(FLAGDOUBLE),
    dta_2(FLAGDOUBLE), dt_A(FLAGDOUBLE), dt_a(FLAGDOUBLE), old_dt, new_dt(0.), argOfLog;
  double a, b, c; // coeffs. in the quadratic eqn ax^2 + bx + c == 0
  double discriminant; // b^2 - 4ac
  long nIter; // number of iterations made while solving a mixed exp-linear equation
  long maxIter(100); // maximum number of iterations allowed

  if (theta_A0 <= 0. || theta_a0 <= 0.)
    {
      string msg = "ActiveLogisticSelectionCoal::PickTime() received an impossible ";
      msg += "pair of Theta values (" + ToString(theta_A0) + " and ";
      msg += ToString(theta_a0) + ").";
      throw impossible_error(msg);
    }

  // Draw the length of the coalescent interval in "magic time,"
  // from an exponential distribution.
  // Do this for the population of lineages with allele A (favored allele),
  // and for the population of lineages with allele a (disfavored allele).

  if (ar->m_xactives[0] > 1)
    {
      // Solve the mixed exp-linear time transformation iteratively.
      // Start by making an approximate guess at the time transformation,
      // by assuming dt is small and expanding (exp(s*t)-1)/s in powers of dt.
      dtau_A = -log(ar->randomSource->Float()) /
	(ar->m_xactives[0]*(ar->m_xactives[0] - 1.)/theta_A0);
      if (fabs(s) > 0.001)
	b = SafeProductWithExp(theta_a0, s*starttime);
      else
	b = theta_a0*(1. + s*starttime); // approximation for small "s"
      a = 0.5 * b * s; // a == 0 if s == 0 or if b underflows (s << 0)
      b += theta_A0; // b > 0
      c = -dtau_A*(theta_A0+theta_a0); // c < 0

      if (b < 0.999*DBL_BIG) // good, no overflow
	{
	  discriminant = b*b - 4.*a*c;
	  if (discriminant >= 0. && 0. != a)
	    {
	      // Because of the values involved, taking the "+" sign
	      // in the quadratic solution is guaranteed to give the desired solution,
	      // for both a > 0 and a < 0.
	      new_dt = (-b + sqrt(discriminant))/(2.*a);
	      if (fabs(s) <= 0.001)
		{
		  dt_A = new_dt; // this solution is "exact"--no need to iterate
		  maxIter = 0; // do not compute anything else for this population
		}
	    }
	  else if (0. == s)
	    {
	      dt_A = dtau_A; // the standard Kingman interval length
	      maxIter = 0; // do not compute anything else for this population
	    }
	  else
	    new_dt = dtau_A * 2.; // use a heuristic guess instead
	  dtA_2 = new_dt; // store this initial guess so we don't recompute it
	}
      else // overflow, s >> 0
	{
	  dt_A = (log((s*(-c))/theta_a0) - s*starttime)/s;
	  maxIter = 0;
	}

      nIter = 0;
      old_dt = 2.*new_dt; // meaningless initialization value that, when appropriate,
                          // guarantees we'll enter the "while" loop
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
	{
	  old_dt = new_dt;
	  //new_dt = (dtau_A*(theta_A0+theta_a0) - theta_a0*exp(s*starttime)*(exp(s*old_dt)-1.)/s)/theta_A0;
	  new_dt = (dtau_A*(theta_A0+theta_a0)-(b-theta_A0)*(SafeProductWithExp(1.,s*old_dt)-1.)/s)/theta_A0;
	  nIter++;
	}
      // note that if new_dt erroneously goes negative, the loop will terminate
      dtA_1 = new_dt;
      if (new_dt > 0. && nIter < maxIter)
	{
	  maxIter = 0; // found a solution; no need to use the other formula
	  new_dt =  -1.; // bogus negative value
	}
      else
	new_dt = dtA_2; // precomputed initial guess
      
      nIter = 0;
      old_dt = 2.*new_dt; // just to ensure the initial "while" condition is met, when appropriate
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
	{
	  old_dt = new_dt;
	  //argOfLog = theta_a0*exp(s*starttime); // compute and test this before dividing by it
	  argOfLog = b - theta_A0; // avoid call to exp() by re-using variable "b"
	  if (argOfLog > 0.) // then it's safe to divide by it
	    argOfLog = 1. + s*(dtau_A*(theta_A0+theta_a0) - theta_A0*old_dt)/argOfLog;
	  if (argOfLog > 0.)
	    new_dt = log(argOfLog)/s;
	  else
	    new_dt = -1.; // this will terminate the loop
	  nIter++;
	}
      // note that if new_dt erroneously goes negative, the loop will terminate
      dtA_2 = new_dt;

      if (fabs(s) > 0.001 && b < 0.999*DBL_BIG)
	{
	  if (dtA_1 > 0.)
	    {
	      if (dtA_2 <= 0.)
		dt_A = dtA_1;
	      else if (dtA_2 > dtA_1)
		{
		  if (0 != maxIter && nIter < maxIter)
		    dt_A = dtA_2; // equation 2 converged, equation 1 didn't
		  else
		    dt_A = dtA_1;
		}
	      else // dtA_1 >= dtA_2 > 0
		dt_A = dtA_2;
	    }
	  else if (dtA_2 > 0.)
	    dt_A = dtA_2;
	  else
	    dt_A = MAX_LENGTH;
	}
      // else dt_A has already been assigned
    }

  if (ar->m_xactives[1] > 1)
    {
      // Solve the mixed exp-linear time transformation iteratively.
      // Start by making an approximate guess at the time transformation,
      // by assuming dt is small and expanding (exp(s*t)-1)/s in powers of dt.
      maxIter = 100; // just in case it got set to 0 above
      dtau_a = -log(ar->randomSource->Float()) /
	(ar->m_xactives[1]*(ar->m_xactives[1] - 1.)/theta_a0);
      if (fabs(s) > 0.001)
	b = SafeProductWithExp(theta_A0, -s*starttime);
      else
	b = theta_A0*(1. - s*starttime); // approximation for small "s"
      a = -0.5 * b * s; // a == 0 if s ==0 or if b underflows (s >> 0)
      b += theta_a0; // b > 0
      c = -dtau_a * (theta_A0 + theta_a0); // c < 0
 
      if (b < 0.999*DBL_BIG) // good, no overflow
	{
	  discriminant = b*b - 4.*a*c;
	  if (discriminant >= 0. && 0. != a) 
	    {
	      // See comment in the related code for ar->m_xactives[0] > 1.
	      new_dt = (-b + sqrt(discriminant))/(2.*a);
	      if (fabs(s) <= 0.001)
		{
		  dt_a = new_dt;
		  maxIter = 0;
		}
	    }
	  else if (0. == s)
	    {
	      dt_a = dtau_a;
	      maxIter = 0;
	    }
	  else
	    new_dt = dtau_a * 2.; // use a heuristic guess instead
	  dta_2 = new_dt; // store this initial guess so we don't recompute it      
	}
      else // overflow, s << 0
	{
	  dt_a = (log((s*c)/theta_A0) + s*starttime)/(-s);
	  maxIter = 0;
	}

      nIter = 0;
      old_dt = 2.*new_dt; // meaningless initialization value that, when appropriate,
                          // guarantees we'll enter the "while" loop
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
      {
	old_dt = new_dt;
	//new_dt = (dtau_a*(theta_A0+theta_a0) + theta_A0*exp(-s*starttime)*(exp(-s*old_dt)-1.)/s)/theta_a0;
	new_dt = (dtau_a*(theta_A0+theta_a0) + (b-theta_a0)*(SafeProductWithExp(1.,-s*old_dt)-1.)/s)/theta_a0;
	nIter++;
      }
      // note that if new_dt erroneously goes negative, the loop will terminate
      dta_1 = new_dt;
      if (new_dt > 0. && nIter < maxIter)
	{
	  maxIter = 0; // found a solution; no need to use the other formula
	  new_dt = -1.; // bogus negative value
	}
      else
	new_dt = dta_2; // precomputed initial guess

      nIter = 0;
      old_dt = 2.*new_dt; // just to ensure the initial "while" condition is met, when appropriate
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
	{
	  old_dt = new_dt;
	  // argOfLog = theta_A0*exp(-s*starttime); // compute and test this before dividing by it
	  argOfLog = b - theta_a0; // avoid call to exp() by re-using variable "b"
	  if (argOfLog > 0.)
	    argOfLog = 1. - s*(dtau_a*(theta_A0+theta_a0) - theta_a0*old_dt)/argOfLog;
	  if (argOfLog > 0.)
	    new_dt = log(argOfLog)/(-s);
	  else
	    new_dt = -1.; // this will terminate the loop
	  nIter++;
	}
      // note that if new_dt erroneously goes negative, the loop will terminate
      dta_2 = new_dt;

      if (fabs(s) > 0.001 && b < 0.999*DBL_BIG)
	{
	  if (dta_1 > 0.)
	    {
	      if (dta_2 <= 0.)
		dt_a = dta_1;
	      else if (dta_2 > dta_1)
		{
		  if (0 != maxIter && nIter < maxIter)
		    dt_a = dta_2; // equation 2 converged, equation 1 didn't
		  else
		    dt_a = dta_1;
		}
	      else // dta_1 >= dta_2 > 0
		dt_a = dta_2;
	    }
	  else if (dta_2 > 0.)
	    dt_a = dta_2;
	  else
	    dt_a = MAX_LENGTH;
	}
      // else dt_a has already been assigned
    }

  if ((dt_A < DBL_EPSILON && dt_A != FLAGDOUBLE) || 
      (dt_a < DBL_EPSILON && dt_a != FLAGDOUBLE))
    {
      string msg = "Bad time length (" + ToString(min(dt_A,dt_a));
      msg += ") proposed in ActiveLogisticSelectionCoal::PickTime().";
      tinypopulation_error e(msg);
      throw e;
    }

  if (FLAGDOUBLE == dt_a)
    {
      m_chosenxpart = (FLAGDOUBLE == dt_A) ? FLAGLONG : 0;
      return dt_A;
    }
  if (FLAGDOUBLE == dt_A)
    {
      m_chosenxpart = 1;
      return dt_a;
    }

  if (dt_A < dt_a)
    {
      m_chosenxpart = 0;
      assert(ar->m_xactives[0] > 1);
      return dt_A;
    }

  m_chosenxpart = 1;
  assert(ar->m_xactives[1] > 1);
  return dt_a;

} /* ActiveLogisticSelectionCoal::PickTime */


//____________________________________________________________________________
//____________________________________________________________________________

InactiveCoal::InactiveCoal(const ForceSummary& fs) 
  : CoalEvent(fs, false),
    m_inv2Theta()
{
} /* InactiveCoal copy constructor */

//____________________________________________________________________________

Event* InactiveCoal::Clone() const
{
  InactiveCoal* event = new InactiveCoal(*this);
  return event;

} /* InactiveCoal::Clone */

//____________________________________________________________________________

void InactiveCoal::InstallParameters(const ForceParameters& starts)
{
  CoalEvent::InstallParameters(starts);

  // We store 2/Theta for speed
  m_inv2Theta = starts.GetRegionalThetas();

  // store 2/Theta
  std::transform(m_inv2Theta.begin(), m_inv2Theta.end(),
		  m_inv2Theta.begin(),
		  std::bind1st(std::divides<double>(),2.0));

} /* InstallParameters */

//____________________________________________________________________________

void InactiveCoal::InstallThetas(const DoubleVec1d& thetas)
{
  Event::InstallThetas(thetas);

  m_inv2Theta = thetas;
  // store 2/Theta
  std::transform(m_inv2Theta.begin(), m_inv2Theta.end(),
		  m_inv2Theta.begin(),
		  std::bind1st(std::divides<double>(),2.0));

} /* InstallThetas */

//____________________________________________________________________________

double InactiveCoal::PickTime(double)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of this "delta t" is theta/(j*k),
  // where j and k (the numbers of inactive and active lineages) and theta
  // correspond to the population which yields the smallest "delta t."

  map<double, long> times;
  double newtime = 0.0;
  long i;

  for (i = 0; i < m_nxparts; ++i) // loop over populations
    {
      if (ar->m_xactives[i] > 0 && ar->m_xinactives[i] > 0)
	{
	  // denominator contains j*k, inactives*actives,
	  // instead of k*(k-1)
	  newtime = - log (ar->randomSource->Float()) /
	    (m_inv2Theta[i] * ar->m_xactives[i] * ar->m_xinactives[i]);
	  times.insert(make_pair(newtime,i));
	}
    }

  if (times.empty())
    {
      // no event is possible
      m_chosenxpart = FLAGLONG;
      return FLAGDOUBLE;
    }
  else
    {
      // the first map entry is the smallest, and thus chosen, time
      map<double, long>::const_iterator mapit = times.begin();
      m_chosenxpart = (*mapit).second;
      assert ((*mapit).first >= 0.0);
      return (*mapit).first;
    }

} /* InactiveCoal::PickTime */

//____________________________________________________________________________
//____________________________________________________________________________

InactiveGrowCoal::InactiveGrowCoal(const ForceSummary& fs) 
 : CoalEvent(fs, false)
{
} /* InactiveGrowCoal constructor */

//____________________________________________________________________________

Event* InactiveGrowCoal::Clone() const
{
  InactiveGrowCoal* event = new InactiveGrowCoal(*this);
  return event;

} /* InactiveGrowCoal::Clone */

//____________________________________________________________________________

double InactiveGrowCoal::PickTime(double starttime)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of this "delta t" is:
  //                           (1/g)*ExpE1(j*k*exp(g*starttime)/(2*g*theta)),
  // where g, theta, j, and k are the growth rate, theta, and numbers of
  // inactive and active lineages for the population which yields the
  // smallest "delta t."  (Note also the factor of 2, absent from ActiveGrowCoal.)
  // For details on the function ExpE1(), see mathx.cpp.

  map<double, long> times;
  double newtime, magicstart;
  long i;

  for (i = 0; i < m_nxparts; ++i) // loop over populations
    {
      if (ar->m_xactives[i] > 0 && ar->m_xinactives[i] > 0)
	{
	  // first do calculation in "magic time"
	  // Numerator contains j*k, inactives*actives,
	  // instead of k*(k-1)
	  // Also, denominator contains a factor of 2, unlike ActiveGrowCoal.
	  newtime = - log(ar->randomSource->Float()) /
	    ((2.0 / m_thetas[i]) * ar->m_xactives[i] * ar->m_xinactives[i]);

	  // then convert to "real time"
	  if (m_growths[i] != 0)
	    {
	      // convert starttime to magical time
	      magicstart = -(1.0 - exp(m_growths[i] * starttime)) / m_growths[i];
	      // add starttime to displacement
	      newtime += magicstart;
	      if (1.0 + m_growths[i] * newtime > 0.0)
		{
		  // convert result to real time
		  newtime = log(1.0 + m_growths[i] * newtime)/m_growths[i];
		  // convert back to displacement
		  newtime -= starttime;
		  // guard against underflow
		  if (newtime < DBL_EPSILON)
		    {
		      tinypopulation_error e("bad time proposed in Event::PickTime");
		      throw e;
		    }
		}
	      else
		newtime = MAX_LENGTH; // make this interval length
	                              // unlikely to win the race
	    }
	  
	  // insert into map (for sorting)
	  times.insert(make_pair(newtime,i));
	}
    }

  if (times.empty())
    {
      // no event is possible
      m_chosenxpart = FLAGLONG;
      return FLAGDOUBLE;
    }
  else
    {
      // the first map entry is the smallest, and thus chosen, time
      map<double, long>::const_iterator mapit = times.begin();
      m_chosenxpart = (*mapit).second;
      assert ((*mapit).first >= 0.0);
      return (*mapit).first;
    }

} /* InactiveGrowCoal::PickTime */


//____________________________________________________________________________
//____________________________________________________________________________

InactiveLogisticSelectionCoal::InactiveLogisticSelectionCoal(const ForceSummary& fs) 
  : CoalEvent(fs, false), m_selectionCoefficient(0.)
{
} /* InactiveLogisticSelectionCoal constructor */

//____________________________________________________________________________

Event* InactiveLogisticSelectionCoal::Clone() const
{
  InactiveLogisticSelectionCoal* event = new InactiveLogisticSelectionCoal(*this);
  return event;

} /* InactiveLogisticSelectionCoal::Clone */

//____________________________________________________________________________

void InactiveLogisticSelectionCoal::InstallParameters(const ForceParameters& starts)
{
  CoalEvent::InstallParameters(starts);
  m_selectionCoefficient = starts.GetLogisticSelectionCoefficient()[0];
  LogisticSelectionTimeSize *pLSTS = dynamic_cast<LogisticSelectionTimeSize*>(Event::m_CalcSizeAtTime);
  if (!pLSTS)
    {
      string msg = "InactiveLogisticSelectionCoal::InstallParameters(), unable ";
      msg += "to dynamically cast its TimeSize pointer to a LogisticSelectionTimeSize ";
      msg += "pointer.";
      throw implementation_error(msg);
    }
  pLSTS->SetSelectionCoefficient(m_selectionCoefficient);
} /* InstallParameters */

//____________________________________________________________________________

double InactiveLogisticSelectionCoal::PickTime(double starttime)
{
  if (2 != m_nxparts)
    {
      string msg = "InactiveLogisticSelectionCoal::PickTime() called with m_nxparts = ";
      msg += ToString(m_nxparts);
      msg += ".  m_nxparts must equal 2, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  if (!((ar->m_xactives[0] >= 1 && ar->m_xinactives[0] >= 1) ||
	(ar->m_xactives[1] >= 1 && ar->m_xinactives[1] >= 1)))
    {
      // no event is possible
      m_chosenxpart = FLAGLONG;
      return FLAGDOUBLE;
    }

  double theta_A0(m_thetas[0]), theta_a0(m_thetas[1]), s(m_selectionCoefficient);
  double dtau_A, dtau_a, dtA_1(FLAGDOUBLE), dtA_2(FLAGDOUBLE), dta_1(FLAGDOUBLE),
    dta_2(FLAGDOUBLE), dt_A(FLAGDOUBLE), dt_a(FLAGDOUBLE), old_dt, new_dt(0.), argOfLog;
  double a, b, c; // coeffs. in the quadratic eqn ax^2 + bx + c == 0
  double discriminant; // b^2 - 4ac
  long nIter; // number of iterations made while solving a mixed exp-linear equation
  long maxIter(100); // maximum number of iterations allowed

  if (theta_A0 <= 0. || theta_a0 <= 0.)
    {
      string msg = "ActiveLogisticSelectionCoal::PickTime() received an impossible ";
      msg += "pair of Theta values (" + ToString(theta_A0) + " and ";
      msg += ToString(theta_a0) + ").";
      throw impossible_error(msg);
    }

  // Draw the length of the coalescent interval in "magic time,"
  // from an exponential distribution.
  // Do this for the population of lineages with allele A (favored allele),
  // and for the population of lineages with allele a (disfavored allele).

  if (ar->m_xactives[0] > 0 && ar->m_xinactives[0] > 0)
    {
      // Solve the mixed exp-linear time transformation iteratively.
      // Start by making an approximate guess at the time transformation,
      // by assuming dt is small and expanding (exp(s*t)-1)/s in powers of dt.
      dtau_A = -log(ar->randomSource->Float()) / 
        (2.*ar->m_xactives[0]*ar->m_xinactives[0]/theta_A0);
      if (fabs(s) > 0.001)
	b = SafeProductWithExp(theta_a0, s*starttime);
      else
	b = theta_a0*(1. + s*starttime); // approximation for small "s"
      a = 0.5 * b * s; // a == 0 if s == 0 or if b underflows (s << 0)
      b += theta_A0; // b > 0
      c = -dtau_A*(theta_A0+theta_a0); // c < 0

      if (b < 0.999*DBL_BIG) // good, no overflow
	{
	  discriminant = b*b - 4.*a*c;
	  if (discriminant >= 0. && 0. != a)
	    {
	      // Because of the values involved, taking the "+" sign
	      // in the quadratic solution is guaranteed to give the desired solution,
	      // for both a > 0 and a < 0.
	      new_dt = (-b + sqrt(discriminant))/(2.*a);
	      if (fabs(s) <= 0.001)
		{
		  dt_A = new_dt; // this solution is "exact"--no need to iterate
		  maxIter = 0; // do not compute anything else for this population
		}
	    }
	  else if (0. == s)
	    {
	      dt_A = dtau_A; // the standard Kingman interval length
	      maxIter = 0; // do not compute anything else for this population
	    }
	  else
	    new_dt = dtau_A * 2.; // use a heuristic guess instead
	  dtA_2 = new_dt; // store this initial guess so we don't recompute it
	}
      else // overflow, s >> 0
	{
	  dt_A = (log((s*(-c))/theta_a0) - s*starttime)/s;
	  maxIter = 0;
	}

      nIter = 0;
      old_dt = 2.*new_dt; // meaningless initialization value that, when appropriate,
                          // guarantees we'll enter the "while" loop
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
	{
	  old_dt = new_dt;
	  //new_dt = (dtau_A*(theta_A0+theta_a0) - theta_a0*exp(s*starttime)*(exp(s*old_dt)-1.)/s)/theta_A0;
	  new_dt = (dtau_A*(theta_A0+theta_a0)-(b-theta_A0)*(SafeProductWithExp(1.,s*old_dt)-1.)/s)/theta_A0;
	  nIter++;
	}
      // note that if new_dt erroneously goes negative, the loop will terminate
      dtA_1 = new_dt;
      if (new_dt > 0. && nIter < maxIter)
	{
	  maxIter = 0; // found a solution; no need to use the other formula
	  new_dt =  -1.; // bogus negative value
	}
      else
	new_dt = dtA_2; // precomputed initial guess
      
      nIter = 0;
      old_dt = 2.*new_dt; // just to ensure the initial "while" condition is met, when appropriate
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
	{
	  old_dt = new_dt;
	  //argOfLog = theta_a0*exp(s*starttime); // compute and test this before dividing by it
	  argOfLog = b - theta_A0; // avoid call to exp() by re-using variable "b"
	  if (argOfLog > 0.) // then it's safe to divide by it
	    argOfLog = 1. + s*(dtau_A*(theta_A0+theta_a0) - theta_A0*old_dt)/argOfLog;
	  if (argOfLog > 0.)
	    new_dt = log(argOfLog)/s;
	  else
	    new_dt = -1.; // this will terminate the loop
	  nIter++;
	}
      // note that if new_dt erroneously goes negative, the loop will terminate
      dtA_2 = new_dt;

      if (fabs(s) > 0.001 && b < 0.999*DBL_BIG)
	{
	  if (dtA_1 > 0.)
	    {
	      if (dtA_2 <= 0.)
		dt_A = dtA_1;
	      else if (dtA_2 > dtA_1)
		{
		  if (0 != maxIter && nIter < maxIter)
		    dt_A = dtA_2; // equation 2 converged, equation 1 didn't
		  else
		    dt_A = dtA_1;
		}
	      else // dtA_1 >= dtA_2 > 0
		dt_A = dtA_2;
	    }
	  else if (dtA_2 > 0.)
	    dt_A = dtA_2;
	  else
	    dt_A = MAX_LENGTH;
	}
      // else dt_A has already been assigned
    }

  if (ar->m_xactives[1] > 0 && ar->m_xinactives[1] > 0) 
    {
      // Solve the mixed exp-linear time transformation iteratively.
      // Start by making an approximate guess at the time transformation,
      // by assuming dt is small and expanding (exp(s*t)-1)/s in powers of dt.
      maxIter = 100; // just in case it got set to 0 above
      dtau_a = -log(ar->randomSource->Float()) / 
        (2.*ar->m_xactives[1]*ar->m_xinactives[1]/theta_a0); 
      if (fabs(s) > 0.001)
	b = SafeProductWithExp(theta_A0, -s*starttime);
      else
	b = theta_A0*(1. - s*starttime); // approximation for small "s"
      a = -0.5 * b * s; // a == 0 if s ==0 or if b underflows (s >> 0)
      b += theta_a0; // b > 0
      c = -dtau_a * (theta_A0 + theta_a0); // c < 0
 
      if (b < 0.999*DBL_BIG) // good, no overflow
	{
	  discriminant = b*b - 4.*a*c;
	  if (discriminant >= 0. && 0. != a) 
	    {
	      // See comment in the related code for ar->m_xactives[0] > 1.
	      new_dt = (-b + sqrt(discriminant))/(2.*a);
	      if (fabs(s) <= 0.001)
		{
		  dt_a = new_dt;
		  maxIter = 0;
		}
	    }
	  else if (0. == s)
	    {
	      dt_a = dtau_a;
	      maxIter = 0;
	    }
	  else
	    new_dt = dtau_a * 2.; // use a heuristic guess instead
	  dta_2 = new_dt; // store this initial guess so we don't recompute it      
	}
      else // overflow, s << 0
	{
	  dt_a = (log((s*c)/theta_A0) + s*starttime)/(-s);
	  maxIter = 0;
	}

      nIter = 0;

      old_dt = 2.*new_dt; // meaningless initialization value that, when appropriate,
                          // guarantees we'll enter the "while" loop
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
      {
	old_dt = new_dt;
	//new_dt = (dtau_a*(theta_A0+theta_a0) + theta_A0*exp(-s*starttime)*(exp(-s*old_dt)-1.)/s)/theta_a0;
	new_dt = (dtau_a*(theta_A0+theta_a0) + (b-theta_a0)*(SafeProductWithExp(1.,-s*old_dt)-1.)/s)/theta_a0;
	nIter++;
      }
      // note that if new_dt erroneously goes negative, the loop will terminate
      dta_1 = new_dt;
      if (new_dt > 0. && nIter < maxIter)
	{
	  maxIter = 0; // found a solution; no need to use the other formula
	  new_dt = -1.; // bogus negative value
	}
      else
	new_dt = dta_2; // precomputed initial guess

      nIter = 0;
      old_dt = 2.*new_dt; // just to ensure the initial "while" condition is met, when appropriate
      while (0. != new_dt && fabs(new_dt - old_dt)/new_dt > 0.00001 && nIter < maxIter)
	{
	  old_dt = new_dt;
	  // argOfLog = theta_A0*exp(-s*starttime); // compute and test this before dividing by it
	  argOfLog = b - theta_a0; // avoid call to exp() by re-using variable "b"
	  if (argOfLog > 0.)
	    argOfLog = 1. - s*(dtau_a*(theta_A0+theta_a0) - theta_a0*old_dt)/argOfLog;
	  if (argOfLog > 0.)
	    new_dt = log(argOfLog)/(-s);
	  else
	    new_dt = -1.; // this will terminate the loop
	  nIter++;
	}
      // note that if new_dt erroneously goes negative, the loop will terminate
      dta_2 = new_dt;

      if (fabs(s) > 0.001 && b < 0.999*DBL_BIG)
	{
	  if (dta_1 > 0.)
	    {
	      if (dta_2 <= 0.)
		dt_a = dta_1;
	      else if (dta_2 > dta_1)
		{
		  if (0 != maxIter && nIter < maxIter)
		    dt_a = dta_2; // equation 2 converged, equation 1 didn't
		  else
		    dt_a = dta_1;
		}
	      else // dta_1 >= dta_2 > 0
		dt_a = dta_2;
	    }
	  else if (dta_2 > 0.)
	    dt_a = dta_2;
	  else
	    dt_a = MAX_LENGTH;
	}
      // else dt_a has already been assigned
    }

  if ((dt_A < DBL_EPSILON && dt_A != FLAGDOUBLE) || 
      (dt_a < DBL_EPSILON && dt_a != FLAGDOUBLE))
    {
      string msg = "Bad time length (" + ToString(min(dt_A,dt_a));
      msg += ") proposed in ActiveLogisticSelectionCoal::PickTime().";
      tinypopulation_error e(msg);
      throw e;
    }

  if (FLAGDOUBLE == dt_a)
    {
      m_chosenxpart = (FLAGDOUBLE == dt_A) ? FLAGLONG : 0;
      return dt_A;
    }
  if (FLAGDOUBLE == dt_A)
    {
      m_chosenxpart = 1;
      return dt_a;
    }

  if (dt_A < dt_a)
    {
      m_chosenxpart = 0;
      assert(ar->m_xactives[0] > 1);
      return dt_A;
    }

  m_chosenxpart = 1;
  assert(ar->m_xactives[1] > 1);
  return dt_a;

} /* InactiveLogisticSelectionCoal::PickTime */

//____________________________________________________________________________
//____________________________________________________________________________

PartitionEvent::PartitionEvent(const ForceSummary& fs)
	: Event(fs)
{
} /* PartitionEvent ctor */

//____________________________________________________________________________
//____________________________________________________________________________

MigEvent::MigEvent(const ForceSummary& fs)
	: PartitionEvent(fs),
	m_rescaledMigRates(),
	m_immigrationRates(),
	m_frompop(FLAGLONG),
	m_topop(FLAGLONG)
{
        m_nparts = registry.GetDataPack().GetNPartitionsByForceType(force_MIG);
	m_forcetype = force_MIG;
        m_maxEvents = fs.GetMaxEvents(force_MIG);
        m_pindex = fs.GetPartIndex(force_MIG);
} /* MigEvent ctor */

//____________________________________________________________________________

Event* MigEvent::Clone() const
{
  MigEvent* event = new MigEvent(*this);
  return event;

} /* MigEvent::Clone */

//____________________________________________________________________________

void MigEvent::InstallParameters(const ForceParameters& starts)
{
  Event::InstallParameters(starts);

  // pre-compute migration rate tables
  ComputeCumulativeRates(starts);

} /* InstallParameters */

//____________________________________________________________________________

void MigEvent::ComputeCumulativeRates(const ForceParameters& starts)
{
  vector<vector<double> > migRates = starts.GetRegional2dRates(force_MIG);
  vector<double> indRate;

  m_rescaledMigRates.clear();
  m_immigrationRates.clear();

  double totalMig;
  long i, j;

  // computes cumulative migration rates in "m_rescaledMigRates"
  // and total immigration into a population in "m_immigrationRates"

  for (i = 0; i < m_nparts; ++i) {
    totalMig = 0.0;
    indRate.clear();
    for (j = 0; j < m_nparts; ++j) {
       totalMig += migRates[i][j];
       indRate.push_back(totalMig);     // cumulative rate
    }

    // normalize cumulative rates to total
    if (totalMig > 0) {
      for (j = 0; j < m_nparts; ++j) {
        indRate[j] /= totalMig;   
      }
    }

    m_rescaledMigRates.push_back(indRate);      // cumulative rates
    m_immigrationRates.push_back(totalMig);     // total immigration
  }

} /* ComputeCumulativeRates */

//____________________________________________________________________________

double MigEvent::PickTime(double)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of this "delta t" is 1/(M*k), where M and k
  // are the immigration rate and number of active lineages of the population
  // which yields the smallest "delta t."

  map<double, long> times;
  double newtime;
  long i;

  for (i = 0; i < m_nparts; ++i)
    {
      if (ar->m_pactives[m_pindex][i] > 0)
	{
	  newtime = -log (ar->randomSource->Float()) /
	    (m_immigrationRates[i] * ar->m_pactives[m_pindex][i]);
	  times.insert(make_pair(newtime, i));
	}  
    }

  if (times.empty())
    {
      // no event is possible
      m_frompop = FLAGLONG;
      m_topop = FLAGLONG;
      return FLAGDOUBLE;
    }
  else
    {
      // the first map entry is the smallest, and thus chosen, time
      map<double, long>::const_iterator mapit = times.begin();
      m_frompop = (*mapit).second;
      double rn = ar->randomSource->Float();
      for (m_topop = 0; m_rescaledMigRates[m_frompop][m_topop] < rn; m_topop++)
	{};
      assert ((*mapit).first >= 0.0);
      return (*mapit).first;
    }

} /* MigEvent::PickTime */

//____________________________________________________________________________

void MigEvent::DoEvent(double eventT)
{
  assert(m_frompop >= 0 && m_topop >= 0);

  // Pick an active branch from the population at random.
  double rn = ar->randomSource->Float();
  Branch_ptr active = ar->m_activelist.RemovePartitionBranch(force_MIG, m_frompop, rn);
  FinishEvent(eventT, active, m_topop);
} /* DoEvent */

void MigEvent::FinishEvent(double eventT, Branch_ptr active, long topop)
{

  Branch_ptr newbranch = ar->GetTree()->Migrate(eventT, topop, m_maxEvents, active);

  ar->m_activelist.Append(newbranch);

} /* FinishEvent */    

//____________________________________________________________________________
//____________________________________________________________________________

DiseaseEvent::DiseaseEvent(const ForceSummary& fs)
	: PartitionEvent(fs),
	m_probhere(),
	m_murateto(),
	m_startdis(FLAGLONG),
	m_enddis(FLAGLONG)
{
 m_nparts = registry.GetDataPack().GetNPartitionsByForceType(force_DISEASE);
	m_forcetype = force_DISEASE;
        m_maxEvents = fs.GetMaxEvents(force_DISEASE);
        m_pindex = fs.GetPartIndex(force_DISEASE);
} /* DiseaseEvent ctor */

//____________________________________________________________________________

Event* DiseaseEvent::Clone() const
{
DiseaseEvent* event = new DiseaseEvent(*this);
return event;

} /* DiseaseEvent::Clone */

//____________________________________________________________________________

double DiseaseEvent::PickTime(double)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of "delta t" is 1/(mu*k), where mu and k
  // are the mutation rate to the disease state and the number of active lineages
  // of the population which yields the smallest value of "delta t."

  map<double, long> times;
  double newtime;
  long i;

  for (i = 0; i < m_nparts; ++i)
    {
      if (ar->m_pactives[m_pindex][i] > 0)
	{
	  newtime = -log (ar->randomSource->Float()) /
	    (m_murateto[i] * ar->m_pactives[m_pindex][i]);
	  times.insert(make_pair(newtime, i));
	}  
    }

  if (times.empty())
    {
      // no event is possible
      m_startdis = FLAGLONG;
      m_enddis = FLAGLONG;
      return FLAGDOUBLE;
    }
  else
    {
      // the first map entry is the smallest, and thus chosen, time
      map<double, long>::const_iterator mapit = times.begin();
      m_startdis = (*mapit).second;
      double rn = ar->randomSource->Float();
      for (m_enddis = 0; m_probhere[m_startdis][m_enddis] < rn;
	   m_enddis++)
	{};
      assert ((*mapit).first >= 0.0);
      return (*mapit).first;
    }

} /* DiseaseEvent::PickTime */

//____________________________________________________________________________

void DiseaseEvent::DoEvent(double eventT)
{
  assert(m_startdis >= 0 && m_enddis >= 0);

  // Pick an active branch from the disease category at random.
  double rn = ar->randomSource->Float();
  Branch_ptr active = ar->m_activelist.RemovePartitionBranch(force_DISEASE, m_startdis, rn);
  FinishEvent(eventT, active, m_enddis);
} /* DoEvent */

void DiseaseEvent::FinishEvent(double eventT, Branch_ptr active, long enddis)
{
  Branch_ptr newbranch = ar->GetTree()->DiseaseMutate(eventT,enddis,m_maxEvents,active);
  ar->m_activelist.Append(newbranch);
} /* DiseaseEvent::FinishEvent */

//____________________________________________________________________________

void DiseaseEvent::InstallParameters(const ForceParameters& starts)
{
  Event::InstallParameters(starts);
  ComputeCumulativeRates(starts);

} /* DiseaseEvent::InstallParameters */

//____________________________________________________________________________

void DiseaseEvent::ComputeCumulativeRates(const ForceParameters& starts)
{
  vector<vector<double> > murates = starts.GetRegional2dRates(force_DISEASE);
  vector<double> indRate;

  m_probhere.clear();
  m_murateto.clear();

  double totaldisease;
  long i, j;

  // computes chance of having a mutation in "probhere"
  // and total rate of change to a disease state in "murateto"


  for (i = 0; i < m_nparts; ++i) {
    totaldisease = 0.0;
    indRate.clear();
    for (j = 0; j < m_nparts; ++j) {
       totaldisease += murates[i][j];
       indRate.push_back(totaldisease);     // cumulative rate
    }

    // normalize cumulative rates to total
    if (totaldisease > 0) {
      for (j = 0; j < registry.GetDataPack().GetNPartitionsByForceType(force_DISEASE); ++j) {
        indRate[j] /= totaldisease;   
      }
    }

    m_probhere.push_back(indRate);      // cumulative rates
    m_murateto.push_back(totaldisease); // total immigration
  }

} /* ComputeCumulativeRates */

//____________________________________________________________________________
//____________________________________________________________________________

ActiveRec::ActiveRec(const ForceSummary& fs)
	: Event(fs),
	m_recrate(defaults::recombinationRate),
	m_onPartitionForces()
{
        m_maxEvents = fs.GetMaxEvents(force_REC);
        m_pindex = FLAGLONG; // we don't use partitions...yet!
} /* ActiveRec ctor */

//____________________________________________________________________________

Event* ActiveRec::Clone() const
{
  ActiveRec* event = new ActiveRec(*this);
  return event;
} /* ActiveRec::Clone */

//____________________________________________________________________________

void ActiveRec::InstallParameters(const ForceParameters& starts)
{
  Event::InstallParameters(starts);

  assert(starts.GetRecRates().size() == 1);  // only one rec rate!
  m_recrate = starts.GetRecRates()[0];

} /* InstallParameters */

//____________________________________________________________________________

double ActiveRec::PickTime(double)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of "delta t" is 1/(r*nA), where r and nA are the
  // recombination rate and number of active links for the population which yields
  // the smallest value of "delta t."

  RecTree* tr = dynamic_cast<RecTree*>(ar->GetTree());
  assert(!(tr->ActiveLinks() > 0 && ar->ActiveSize() == 0));  
  // active sites w/o lineages
  if (tr->ActiveLinks() > 0)
    {
    double retval = -log(ar->randomSource->Float()) / (m_recrate * tr->ActiveLinks());
      assert (retval >= 0.0);
      return retval;
    }
  else
    return FLAGDOUBLE;

} /* ActiveRec::PickTime */

//____________________________________________________________________________

void ActiveRec::DoEvent(double eventT)
{
  RecTree* tr = dynamic_cast<RecTree*>(ar->GetTree());

  // Pick an active branch and recombination site at random.
  Branch_ptr active;
  long rn = ar->randomSource->Long(tr->ActiveLinks());
  Branchiter brit;
  for (brit = ar->m_activelist.Begin(); ; ++brit)
  {
    active = *brit;
    long n = active->m_range.ActiveLinks();
    if (n > rn) break;   // found it!
    rn -= n;
  }
  long recsite = active->m_range.GetFirstActivesite() + rn;
  ar->m_activelist.Remove(brit);

  //Pick a partition at random if we have a local partition force (i.e.
  // Disease) on and active at this region.
  FPartMap fparts;
  if (!m_onPartitionForces.empty()) {
    for(vector<force_type>::iterator spart = m_onPartitionForces.begin();
        spart != m_onPartitionForces.end(); ++spart) {
      if (*spart != force_MIG) {
        DoubleVec1d allSizes = m_CalcSizeAtTime->
          PartitionSizeAt(*spart,m_thetas,m_growths,eventT);
        long randomPart = ChooseRandomFromWeights(allSizes);
        fparts.insert(make_pair(*spart, randomPart));
      }
    }
  }
  
  FinishEvent(eventT, active, recsite, fparts);
} /* DoEvent */

branchpair ActiveRec::FinishEvent(double eventT, Branch_ptr active, long recsite,
                            FPartMap fparts)
{
  RecTree* tr = dynamic_cast<RecTree*>(ar->GetTree());

  branchpair newbranches = 
     tr->RecombineActive(eventT, m_maxEvents, fparts, active, recsite); 

  ar->m_activelist.Append(newbranches.first);
  ar->m_activelist.Append(newbranches.second);
  assert(newbranches.first->PartitionsConsistentWith(active));
  assert(newbranches.second->PartitionsConsistentWith(active));
  return newbranches;

} /* FinishEvent */
    
//____________________________________________________________________________
//____________________________________________________________________________

InactiveRec::InactiveRec(const ForceSummary& fs)
	: Event(fs),
          m_recrate(defaults::recombinationRate)
{
          m_maxEvents = fs.GetMaxEvents(force_REC);
          m_pindex = FLAGLONG; // we don't use partitions...yet!
} /* InactiveRec ctor */

//____________________________________________________________________________

Event* InactiveRec::Clone() const
{
  InactiveRec* event = new InactiveRec(*this);
  return event;

} /* InactiveRec::Clone */

//____________________________________________________________________________

void InactiveRec::InstallParameters(const ForceParameters& starts)
{
  Event::InstallParameters(starts);

  assert(starts.GetRecRates().size() == 1); // only one rec rate!
  m_recrate = starts.GetRecRates()[0];

} /* InstallParameters */

//____________________________________________________________________________

double InactiveRec::PickTime(double)
{
  // Computes a displacement, a.k.a. "delta t" -- the length of the time interval,
  // not the timestamp for the end of the interval.

  // Note:  The expectation value of "delta t" is 1/(r*nO), where r and nO are the
  // recombination rate and number of open links of the population which yields
  // the smallest value of "delta t."

  RecTree* tr = dynamic_cast<RecTree*>(ar->GetTree());
  long links = tr->OpenLinks();
  if (links > 0)
    {
      double retval = -log(ar->randomSource->Float()) / (m_recrate * links);
      assert(retval >= 0.0);
      return retval;
    }
  else
    return FLAGDOUBLE;

} /* InactiveRec::PickTime */

//____________________________________________________________________________

void InactiveRec::DoEvent(double eventT)
{
  RecTree* tr = dynamic_cast<RecTree*>(ar->GetTree());

  //Pick a partition at random if we have a local partition force (i.e.
  // Disease) on and active at this region.
  FPartMap fparts;
  if (!m_onPartitionForces.empty()) {
    for(vector<force_type>::iterator spart = m_onPartitionForces.begin();
        spart != m_onPartitionForces.end(); ++spart) {
      if (*spart != force_MIG) {
        DoubleVec1d allSizes = m_CalcSizeAtTime->
          PartitionSizeAt(*spart,m_thetas,m_growths,eventT);
        long randomPart = ChooseRandomFromWeights(allSizes);
        fparts.insert(make_pair(*spart, randomPart));
      }
    }
  }

  // Pick an open branch and recombination site at random.
  Branch_ptr branch;
  long rn = ar->randomSource->Long(tr->OpenLinks());
  Branchiter brit;
  for (brit = ar->m_inactivelist.Begin(); ; ++brit)
  {
    branch = *brit;
    long n = branch->m_range.NewActiveLinks();
    if (n > rn) break;      // found it!
    rn -= n;
  }

  // at this point "rn" contains the index, counting from 0, of the
  // desired site on branch pBranch

  assert(branch != Branch::NONBRANCH);  // we didn't find a branch?

  ar->m_inactivelist.Remove(brit);

  // Find the recombination site.
  long recsite(FLAGLONG);
  rangeset opensites = branch->m_range.GetNewactivesites();
  rangeset::iterator rit;
  for(rit = opensites.begin(); rit != opensites.end(); ++rit) {
     long s1 = rit->first;
     long s2 = rit->second;
     recsite = s1 + rn;
     if (recsite < s2) break;   // found it!
     rn -= (s2 - s1);
  }

  assert(recsite != FLAGLONG); // failed to find a recombination site!
  branchpair branches = 
     tr->RecombineInactive(eventT, m_maxEvents, fparts, branch, recsite);

  ar->m_inactivelist.Collate(branches.first);
  ar->m_activelist.Append(branches.second);
  assert(branches.first->PartitionsConsistentWith(branch));
  assert(branches.second->PartitionsConsistentWith(branch));
} /* DoEvent */
    
//____________________________________________________________________________
