// $Id: plforces.cpp,v 1.55 2007/04/30 21:50:50 erynes Exp $
//
// Posterior Likelihood forces classes
// - CoalescePL
//   - coalesce with no-growth [using compressed summaries]
// - CoalesceGrowPL
//   - coalesce with exponential growth [using non-compressed summaries]
// - GrowPL
//   - coalesce with exponential growth [using non-compressed summaries]
// - MigratePL
// - RecombinePL
// - SelectPL [stubs]
///////////////////////////////////////////////////////    
/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "mathx.h"
#include "plforces.h"
#include "runreport.h"
#include "summary.h"
#include "vectorx.h"

#ifdef DMALLOC_FUNC_CHECK
#include <dmalloc.h>
#endif

using namespace std;


// DEBUG CODE HELPERS
#include <iostream>
#include <fstream>
#include <sstream>
std::ofstream numfile;

//-----------------------------------------------------
// DiseasePL: disease force specific waiting time and 
//            point probabilites plus derivatives
//-----------------------------------------------------

double
DiseasePL::lnWait (const vector < double >&param, 
		   const TreeSummary * treedata)
{
  vector < double >::iterator i;
  vector < double >::const_iterator pstart;
  vector < double >::const_iterator pend;

  const vector<double>& mWait = treedata->GetDiseaseSummary()->GetShortWait();
  // precomputing the transition rates to a status Sum[d_ji] into
  // membervariable m_msum
  for (i = m_msum.begin (),
       pstart = param.begin () + m_start,
       pend = param.begin () + m_start + m_nPop;
       i != m_msum.end (); ++i, pstart += m_nPop, pend += m_nPop)
    {
      (*i) = accumulate (pstart, pend, 0.0);
    }

  // sum_pop(SM_status *kt)
  return  -1.0 * inner_product (mWait.begin (),
			mWait.end (), m_msum.begin (), 0.0);
}

double
DiseasePL::lnPoint (const vector < double >&param, 
		    const vector < double >&lparam,
		    const TreeSummary * treedata)
{
  //  Sum_j(Sum_i(disevent[j,i] * log(M[j,i]))
  const vector<double>& nevents = treedata->GetDiseaseSummary()->GetShortPoint();
  return  inner_product (nevents.begin (), nevents.end (), 
                        lparam.begin () + m_start, 0.0);
}

double
DiseasePL::DlnWait (const vector < double >&param,
		    const TreeSummary * treedata, const long &whichparam)
{
  long which = (whichparam - m_start) / m_nPop;
  const vector<double>& mWait = treedata->GetDiseaseSummary()->GetShortWait();
  return -mWait[which];
}

double
DiseasePL::DlnPoint (const vector < double >&param,
		     const TreeSummary * treedata, const long &whichparam)
{
  long which = whichparam - m_start;
  const vector<double>& nmig = treedata->GetDiseaseSummary()->GetShortPoint();
  return SafeDivide (nmig[which], param[whichparam]);
}


// MigratePL: migration forces specific waiting time and 
//            point probabilites and its derivatives
double
MigratePL::lnWait (const vector < double >&param, 
		   const TreeSummary * treedata)
{
  vector < double >::iterator i;
  vector < double >::const_iterator pstart;
  vector < double >::const_iterator pend;

  const vector<double>& mWait = treedata->GetMigSummary()->GetShortWait();

  // precomputing the immigration rates Sum[m_ji] into membervariable m_msum
  for (i = m_msum.begin (),
       pstart = param.begin () + m_start,
       pend = param.begin () + m_start + m_nPop;
       i != m_msum.end (); ++i, pstart += m_nPop, pend += m_nPop)
    {
      (*i) = accumulate (pstart, pend, 0.0);
    }

  // sum_pop(SM_pop *kt)
  return  -1.0 * inner_product (mWait.begin (),
  		mWait.end (), m_msum.begin (), 0.0);
}

double
MigratePL::lnPoint (const vector < double >&param, 
		    const vector < double >&lparam,
		    const TreeSummary * treedata)
{
  //  Sum_j(Sum_i(migevent[j,i] * log(M[j,i]))
  const vector<double>& nmig = treedata->GetMigSummary()->GetShortPoint();

  return  inner_product (nmig.begin(), nmig.end(), lparam.begin() + m_start,
  		0.0);
}

double
MigratePL::DlnWait (const vector < double >&param,
		    const TreeSummary * treedata, const long &whichparam)
{
  long which = (whichparam - m_start) / m_nPop;
  const vector<double>& mWait = treedata->GetMigSummary()->GetShortWait();
  return -mWait[which];
}

double
MigratePL::DlnPoint (const vector < double >&param,
		     const TreeSummary * treedata, const long &whichparam)
{
  long which = whichparam - m_start;
  const vector<double>& nmig = treedata->GetMigSummary()->GetShortPoint();

  return SafeDivide (nmig[which], param[whichparam]);
}

//-----------------------------------------------------
// CoalescePL: coalecsence forces specific waiting time and 
//             point probabilites and its derivatives
// COMPRESSED DATA (SHORT) SUMMARIES
double
CoalescePL::lnWait (const vector < double >&param, 
		    const TreeSummary * treedata)
{
  // result = sum_pop(k(k-1)t/theta_pop)
  const vector<double>& cWait = treedata->GetCoalSummary()->GetShortWait();

  return -1.0 * inner_product (cWait.begin (), cWait.end (),
  			param.begin () + m_start,
  			0.0, plus < double >(), 
  			divides < double >());

}

double
CoalescePL::lnPoint (const vector < double >&param, 
		     const vector < double >&lparam,
		     const TreeSummary * treedata)
{
  //  in general it is Sum_j(coalesceevent[j] * (log(2) - log(theta_j)))
  //  sumcpoint = Sum_j(coalesceevent[j] * log(2)) 
  //   - Sum_j(coalesceevent[j] * log(theta_j)) 
  const vector<double>& ncoal = treedata->GetCoalSummary()->GetShortPoint();

  return (LOG2 *
  	  accumulate (ncoal.begin (), ncoal.end (), 0.0) - 
            inner_product (ncoal.begin (), ncoal.end (), 
  			   lparam.begin () + m_start, 0.0));
}

double
CoalescePL::DlnWait (const vector < double >&param,
		     const TreeSummary * treedata, 
		     const long &whichparam)
{
    long which = whichparam - m_start;
    const vector<double>& cWait = treedata->GetCoalSummary()->GetShortWait();

    return cWait[which] / (param[whichparam] * param[whichparam]);
}

// erynes 2004/03/11 -- This method presently returns the following:
//
//          -ncoal/theta,
//
// where theta is the current estimate of parameter theta for the
// given population ("whichparam"), and ncoal is the total number
// of coalescent events for this population (i.e., one less than the
// total number of tips that the population has in this genealogy tree,
// unless recombinations are present).
//
// IMPORTANT NOTE:
// Wait() and Point() return log(WaitProb(G|P)) and log(PointProb(G|P)),
// and PostLike::DCalculate() currently assumes that DWait() and DPoint()
// return the derivatives of log(WaitProb(G|P)) and log(PointProb(G|P)),
// instead of returning the derivatives of WaitProb/PointProb directly.
// As long as everything is kept consistent, this is all very good
// for reducing complexity and improving speed.
//
// Note:  CoalescePL::DPoint calculates the partial derivative of Point
// with respect to theta for the given population ("whichparam"),
// in both the absence and presence of growth.
// When growth is present, CoalesceGrowPL::DWait does the same for Wait.
// Partial derivatives with respect to growth are found in GrowPL.
//
double
CoalescePL::DlnPoint (const vector < double >&param,
		      const TreeSummary * treedata, const long &whichparam)
{
  long which = whichparam - m_start;
  const vector<double>& ncoal = treedata->GetCoalSummary()->GetShortPoint();

  return -ncoal[which] / param[whichparam];
}

//-----------------------------------------------------
// CoalesceGrowPL: coalescence forces specific waiting time and 
//             point probabilites and its derivatives
// LONG DATA SUMMARIES
// [only in effect when GROWTH force is turned on]



// erynes 2004/03/11 -- This method presently returns the following:
//
// sum_over_populations(sum_over_each_population's_coalescent_time_intervals(
//  (k_i*(k_i - 1)/(theta_p * g_p)) * (exp(g_p * t_start)- exp(g_p * t_end))
//                                                                          )
//                     )
//
// where
//        k_i is the number of lineages of population p in time interval i
//        theta_p is the present estimate of parameter theta for population p
//        g_p is the present estimate of the growth parameter for population p
//        t_start is the timepoint at the start of time interval i
//        t_end is the timepoint at the end of time interval i
//
// This quantity is equal to log(WaitProb(G|P)), where WaitProb(G|P) is the
// "waiting" probability of the input genealogy G (parameter "treedata")
// given the parameters P (parameter "param").
// Prob(G|P) = PointProb(G|P) * WaitProb(G|P).
// See CoalesceGrowPL::Point() for the point probability term.
//
double
CoalesceGrowPL::lnWait (const vector < double >&param, 
			const TreeSummary * treedata)
{
  const list<Interval>& treesum=treedata->GetCoalSummary()->GetLongWait();
  list <Interval> :: const_iterator tit;
  vector < double > :: const_iterator pit = param.begin() + m_start;
  vector < double > :: const_iterator pend = param.begin() + m_start + m_nPop;
  vector < double > :: const_iterator git = param.begin() + m_growthstart;

  unsigned long xpartition = 0L;
  double growth, theta, gts, gte;
  double coeff_s, coeff_e, arg1_s, arg1_e;
  double result = 0.0;
  double starttime;

  for( ; pit != pend; ++pit, ++git, ++xpartition) {
     growth = *git;
     theta = *pit;

     if (fabs(growth) < GROWTH_EPSILON)
       {
	 result += lnWaitForTinyGrowth(theta, growth, xpartition, treesum);
	 continue;
       }

     starttime = 0.0;
     for(tit = treesum.begin(); tit != treesum.end(); ++tit) {

       gts = growth * starttime;
       gte = growth * (*tit).endtime;

       coeff_s = coeff_e = arg1_s = arg1_e = 1.;
       const double& k = (*tit).xpartlines[xpartition];

       // IMPORTANT NOTE:  The following over/underflow checking
       // assumes that each starttime and endtime is always >= 0.
       // Hence, e.g., gts < 0 means growth < 0.

       if (gts > 1.)
	 {
	   coeff_s = k*(k-1.);
	   if (growth > 1.)
	     arg1_s /= growth;
	   else
	     coeff_s /= growth;
	   if (theta > 1.)
	     arg1_s /= theta;
	   else
	     coeff_s /= theta;
	 }
       else if (gts < -1.)
	 {
	   arg1_s *= k*(k-1.);
	   if (growth < -1.)
	     coeff_s /= growth;
	   else
	     arg1_s /= growth;
	   if (theta > 1.)
	     coeff_s /= theta;
	   else
	     arg1_s /= theta;
	 }
       else // no over/underflow in exp()
	 coeff_s = k*(k-1.)/(theta*growth); // arg1_s already set

       // BUGBUG erynes -- what follows is easy to understand,
       // but we should revisit this and reuse some of what we
       // learned about growth and theta above, while considering gts.
       
       if (gte > 1.)
	 {
	   coeff_e = k*(k-1.);
	   if (growth > 1.)
	     arg1_e /= growth;
	   else
	     coeff_e /= growth;
	   if (theta > 1.)
	     arg1_e /= theta;
	   else
	     coeff_e /= theta;
	 }
       else if (gte < -1.)
	 {
	   arg1_e *= k*(k-1.);
	   if (growth < -1.)
	     coeff_e /= growth;
	   else
	     arg1_e /= growth;
	   if (theta > 1.)
	     coeff_e /= theta;
	   else
	     arg1_e /= theta;
	 }
       else // no over/underflow in exp()
	 coeff_e = k*(k-1.)/(theta*growth); // arg1_e already set

       double incrementalResult =
	 coeff_s * SafeProductWithExp(arg1_s, gts) -
         coeff_e * SafeProductWithExp(arg1_e, gte);

       if (incrementalResult > 0.)
	 {
	   // We have precision problems,
	   // because we expect te > ts always,
	   // and hence gte > gts always.
	   // During debugging, we have encountered precision problems
	   // in which gdb claims gts == gte and arg1_s == arg1_e
	   // but incrementalResult > 0, typically reflecting
	   // a difference near the 15th digit.
	   // (Note that te == ts corresponds to two events
	   // occurring simultaneously--e.g., k = 1 for a really long
	   // time in each of two populations, then one finally
	   // migrates to the other, then they coalesce instantaneously.)
	   // It should be safe to set incrementalResult to zero in cases of
	   // roundoff error.  If incrementalResult is positive,
	   // but _not_ tiny compared with either of the terms
	   // whose difference yielded incrementalResult,
	   // then something is _very_ wrong--probably te < ts,
	   // which should never happen.  We assert in this case;
	   // this will enable us to see the problem, but the end user
	   // will not see the assertion, unless s/he's running in debug mode.
	   if (incrementalResult/fabs(coeff_s * SafeProductWithExp(arg1_s, gts))
	       > 1e-10) {
             string msg = "incrementalResult in plforces.cpp is greater "
               "than zero.  coeff_s = "
               + ToString(coeff_s) + ", coeff_e = " + ToString(coeff_e)
               + ", theta = " + ToString(theta) + ", growth = "
               + ToString(growth) + ", and incrementalResult is "
               + ToString(incrementalResult);
             registry.GetRunReport().ReportDebug(msg);
	     //assert(0);
           }
	   incrementalResult = 0.;
	 }
	   
       starttime = (*tit).endtime;
       result += incrementalResult;

    }

  }

  return result;
}

// erynes 2004/05/20 -- This method presently returns the following:
//
// sum_over_a_population's_coalescent_time_intervals(
//  (k_i*(k_i - 1)/theta_p) * ( (tis - tie) + (1/2)*g_p*(tis^2 - tie^2) )
//                                                  )
//
// where
//        k_i is the number of lineages of population p in time interval i
//        theta_p is the present estimate of parameter theta for population p
//        g_p is the present estimate of the growth parameter for population p
//        tis is the timepoint at the start of time interval i
//        tie is the timepoint at the end of time interval i
//
// The population "p" is determined in Wait() by iterating over all
// populations.  That method calls this method with the appropriate
// parameters for the population in question.
//
// This non-public method is used to handle the case in which fabs(g)
// is tiny--that is, less than GROWTH_EPSILON.  It is derived from the
// formula used in Wait by expanding exp(g*t) in a Taylor series
// about g == 0, multiplying this by the term k(k-1)/(g*theta),
// and dropping all terms of order g^2 and above.
// Brief tests with Mathematica suggest that the original formula
// converges very nicely to this linear formula for normal values
// of the timepoints, the difference being less than one part in one
// billion, dropping down to zero difference for g == 0 (using
// L'Hopital's rule).
//
// Please also see the comments for DWait.
//
double
CoalesceGrowPL::lnWaitForTinyGrowth(const double theta, const double growth,
				    const unsigned long xpartition,
				    const list<Interval>& treesummary)
{
  assert(fabs(growth) < GROWTH_EPSILON); // otherwise use Wait

  list<Interval>::const_iterator tit;
  double k, te;

  double result = 0.0;
  // ts starts at 0 and subsequently is the previous cycle's te
  double ts = 0.0;

  for(tit = treesummary.begin(); tit != treesummary.end(); ++tit)
    {
       te = (*tit).endtime;
       k  = (*tit).xpartlines[xpartition];

       result += (k*(k-1.)/theta) *
	         ((ts - te) + 0.5*growth*(ts*ts - te*te));
       ts = te;
    }

  return result;
}




// erynes 2004/03/11 -- This method presently returns the following:
//
// sum_over_populations(sum_over_each_population's_coalescent_time_intervals(
//                              log(2) + g_p*t_i_end - log(theta_p)
//                                                                          )
//                     )
//
// where
//        theta_p is the present estimate of parameter theta for population p
//        g_p is the present estimate of the growth parameter for population p
//        t_i_end is the timepoint at the end of time interval i
//
// This quantity is equal to log(PointProb(G|P)), where PointProb(G|P) is the
// "point" probability of the input genealogy G (parameter "treedata")
// given the parameters P (parameter "param").
// Prob(G|P) = PointProb(G|P) * WaitProb(G|P).
// See CoalesceGrowPL::lnWait() for the wait probability term.
//
double
CoalesceGrowPL::lnPoint (const vector < double >&param, 
			 const vector < double >&lparam,
			 const TreeSummary * treedata)
{
  const Interval *  treesum = treedata->GetCoalSummary()->GetLongPoint();
  const Interval * tit;
  vector < double > :: const_iterator pit = lparam.begin() + m_start;
  vector < double > :: const_iterator pend = lparam.begin () + m_start + m_nPop;
  vector < double > :: const_iterator git = param.begin() + m_growthstart;

  long xpartition = 0L;
  double result = 0.0;

  for( ; pit != pend; ++pit, ++git, ++xpartition)
    {
      for(tit= treesum; tit != NULL; tit = tit->next)
	{
	  if(xpartition==(*tit).oldstatus)
	    result +=  LOG2 + (*tit).endtime * (*git) - (*pit);
	}
    }

  return result;
}

// erynes 2004/03/11 -- This method currently returns the following:
//
// sum_over_coalescent_time_intervals_of_population_whichparam(
//   -(k_i(k_i - 1)/(g*theta*theta))*(exp(g*t_i_start) - exp(g*t_i_end))
//                                                            )
// where
//   k_i is the number of lineages of this population in time interval i
//   g   is the current estimate of the growth parameter for this pop.
//   theta is the current estimate of the theta parameter for this pop.
//   t_i_start is the timepoint at the beginning of time interval i
//   t_i_end is the timepoint at the end of time interval i
//
// IMPORTANT NOTE:
// Wait() and Point() return log(WaitProb(G|P)) and log(PointProb(G|P)),
// and PostLike::DCalculate() currently assumes that DWait() and DPoint()
// return the derivatives of log(WaitProb(G|P)) and log(PointProb(G|P)),
// instead of returning the derivatives of WaitProb/PointProb directly.
// As long as everything is kept consistent, this is all very good
// for reducing complexity and improving speed.
//
// Note:  CoalesceGrowPL::DWait calculates the partial derivative of Wait
// with respect to theta for the given population ("whichparam").
// CoalescePL::DPoint does the same for Point.
// Partial derivatives with respect to growth are found in GrowPL.
//
double
CoalesceGrowPL::DlnWait (const vector < double >&param,
			 const TreeSummary * treedata, 
			 const long &whichparam)
{
  const double growth = param[whichparam + m_growthstart];
  if(fabs(growth) < GROWTH_EPSILON)
    return DlnWaitForTinyGrowth(param, treedata, whichparam);

  double result = 0.0;
  long which = whichparam - m_start;
  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list <Interval> :: const_iterator tit=treesum.begin();
  const double theta = param[whichparam];
  double coeff_s, coeff_e, arg1_s, arg1_e;
  double starttime = 0.0;

  for( ; tit != treesum.end(); ++tit)
    {
      double gts = growth * starttime;
      double gte = growth * (*tit).endtime;

      const long& k = (*tit).xpartlines[which];

      coeff_s = coeff_e = arg1_s = arg1_e = 1.;
      
      // IMPORTANT NOTE:  The following over/underflow checking
      // assumes that each starttime and endtime is always >= 0.
      // Hence, e.g., gts < 0 means growth < 0.

      if (gts > 1.)
	{
	  coeff_s = -k*(k-1.);
	  if (growth > 1.)
	    arg1_s /= growth;
	  else
	    coeff_s /= growth;
	  if (theta > 1.)
	    arg1_s /= theta*theta;
	  else
	    coeff_s /= theta*theta;
	  }
      else if (gts < -1.)
	{
	  arg1_s *= -k*(k-1.);
	  if (growth < -1.)
	    coeff_s /= growth;
	  else
	    arg1_s /= growth;
	  if (theta > 1.)
	    coeff_s /= theta*theta;
	  else
	    arg1_s /= theta*theta;
	}
      else // no over/underflow in exp()
	coeff_s = -k*(k-1.)/(theta*theta*growth); // arg1_s already set

      // BUGBUG erynes -- what follows is easy to understand,
      // but we should revisit this and reuse some of what we
      // learned about growth and theta above, while considering gts.

      if (gte > 1.)
	{
	  coeff_e = -k*(k-1.);
	  if (growth > 1.)
	    arg1_e /= growth;
	  else
	    coeff_e /= growth;
	  if (theta > 1.)
	    arg1_e /= theta*theta;
	  else
	    coeff_e /= theta*theta;
	  }
      else if (gte < -1.)
	{
	  arg1_e *= -k*(k-1.);
	  if (growth < -1.)
	    coeff_e /= growth;
	  else
	    arg1_e /= growth;
	  if (theta > 1.)
	    coeff_e /= theta*theta;
	  else
	    arg1_e /= theta*theta;
	}
      else // no over/underflow in exp()
	coeff_e = -k*(k-1.)/(theta*theta*growth); // arg1_e already set

      double incrementalResult =
	coeff_s * SafeProductWithExp(arg1_s, gts) -
	coeff_e * SafeProductWithExp(arg1_e, gte);

      if (incrementalResult < 0.)
	{
	  // We have precision problems,
	  // because we expect te > ts always,
	  // and hence gte > gts always.
	  // During debugging, we have encountered precision problems
	  // in which gdb claims gts == gte and arg1_s == arg1_e
	  // but incrementalResult < 0, typically reflecting
	  // a difference near the 15th digit.
	  // (Note that te == ts corresponds to two events
	  // occurring simultaneously--e.g., k = 1 for a really long
	  // time in each of two populations, then one finally
	  // migrates to the other, then they coalesce instantaneously.)
	  // It should be safe to set incrementalResult to zero in cases of
	  // roundoff error.  If incrementalResult is negative,
	  // but _not_ tiny compared with either of the terms
	  // whose difference yielded incrementalResult,
	  // then something is _very_ wrong--probably te < ts,
	  // which should never happen.  We assert in this case;
	  // this will enable us to see the problem, but the end user
	  // will not see the assertion, unless s/he's running in debug mode.
	  if (incrementalResult/fabs(coeff_s * SafeProductWithExp(arg1_s, gts))
	      < -1e-10)
	    assert(0);
	  incrementalResult = 0.;
	}

      starttime = (*tit).endtime;
      result += incrementalResult;

    }

  return result;
}

// erynes 2004/05/20 -- This method currently returns the following:
//
// sum_over_coalescent_time_intervals_of_population_whichparam(
//   -(k_i(k_i - 1)/(theta*theta))*( (t_i_start - t_i_end) +
//                                   (1/2)*g*(t_i_start^2 - t_i_end^2) )
//                                                            )
// where
//   k_i is the number of lineages of this population in time interval i
//   g   is the current estimate of the growth parameter for this pop.
//   theta is the current estimate of the theta parameter for this pop.
//   t_i_start is the timepoint at the beginning of time interval i
//   t_i_end is the timepoint at the end of time interval i
//
// This non-public method is used to handle the case in which fabs(g)
// is tiny--that is, less than GROWTH_EPSILON.  It is derived from the
// formula used in DWait by expanding exp(g*t) in a Taylor series
// about g == 0, multiplying this by the term -k(k-1)/(g*theta*theta),
// and dropping all terms of order g^2 and above.
// Brief tests with Mathematica suggest that the original formula
// converges very nicely to this linear formula for normal values
// of the timepoints, the difference being less than one part in one
// billion, dropping down to zero difference for g == 0 (using
// L'Hopital's rule).
//
// Please also see the comments for DWait.
//
double
CoalesceGrowPL::DlnWaitForTinyGrowth(const vector < double >&param,
				     const TreeSummary * treedata,
				     const long &whichparam)
{
  const double growth = param[whichparam + m_growthstart];
  assert(fabs(growth) < GROWTH_EPSILON); // otherwise use DWait()
  const double theta = param[whichparam];
  double k, te;
  long which = whichparam - m_start;
  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator tit = treesum.begin();

  double result = 0.0;
  double ts = 0.0;

  for( ; tit != treesum.end(); ++tit)
    {
      k  = (*tit).xpartlines[which];
      te = (*tit).endtime;

      result += -(k*(k-1.)/(theta*theta)) *
	        ((ts - te) + 0.5*growth*(ts*ts - te*te));
      ts = te;
    }

  return result;
}

//-----------------------------------------------------
// GrowPL: exponential growth forces specific waiting time (=0) and 
//              point probabilites and its derivatives
// LONG DATA SUMMARIES
double
GrowPL::lnWait (const vector < double >&param, 
		const TreeSummary * treedata)
{
  // Growth is a modifier of theta, hence its contribution
  // is included in CoalesceGrowPL::Wait.
  return 0.;
}

double
GrowPL::lnPoint (const vector < double >&param, 
		 const vector < double >&lparam,
		 const TreeSummary * treedata)
{
  // Growth is a modifier of theta, hence its contribution
  // is included in CoalesceGrowPL::Point.
  return 0.;
}

// erynes 2004/04/09 -- This method currently returns the following:
//
// sum_over_coalescent_time_intervals_of_population_whichparam(
//   -(k_i(k_i - 1)/(g*g*theta))*(
//             (1 - g*t_i_start)*exp(g*t_i_start) -
//             (1 - g*t_i_end)  *exp(g*t_i_end)
//                               )
//                                                            )
// where
//   k_i is the number of lineages of this population in time interval i
//   g   is the current estimate of the growth parameter for this pop.
//   theta is the current estimate of the theta parameter for this pop.
//   t_i_start is the timepoint at the beginning of time interval i
//   t_i_end is the timepoint at the end of time interval i
//
double
GrowPL::DlnWait (const vector < double >&param,
		 const TreeSummary * treedata, 
		 const long &whichparam)
{
  const double growth = param[whichparam];
  if (fabs(growth) < GROWTH_EPSILON)
    return DlnWaitForTinyGrowth(param, treedata, whichparam);

  const long whicht = m_thetastart + whichparam - m_start;
  const double theta = param[whicht];
  assert(theta); // we divide by it below
  double gte, gts, arg1_s, arg1_e, coeff_s, coeff_e;
  long k;

  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list <Interval> :: const_iterator tit = treesum.begin(); 

  double result = 0.0;
  double starttime = 0.0;

  for( ; tit != treesum.end(); ++tit)
    {
      gts = growth * starttime;
      gte = growth * (*tit).endtime;

      k = (*tit).xpartlines[whicht];

      arg1_s = 1. - gts;
      arg1_e = 1. - gte;
      coeff_s = 1.;
      coeff_e = 1.;

      // IMPORTANT NOTE:  The following over/underflow checking
      // assumes that each starttime and endtime is always >= 0.
      // Hence, e.g., gts < 0 means growth < 0.

      if (gts > 1.)
	{
	  coeff_s = -k*(k-1.);
	  if (growth > 1.)
	    arg1_s /= growth*growth;
	  else
	    coeff_s /= growth*growth;
	  if (theta > 1.)
	    arg1_s /= theta;
	  else
	    coeff_s /= theta;
	}
      else if (gts < -1.)
	{
	  arg1_s *= -k*(k-1.);
	  if (growth < -1.)
	    coeff_s /= growth*growth;
	  else
	    arg1_s /= growth*growth;
	  if (theta > 1.)
	    coeff_s /= theta;
	  else
	    arg1_s /= theta;
	}
      else // no over/underflow in exp()
	coeff_s = -k*(k-1.)/(theta*growth*growth); // arg1_s already set

      // BUGBUG erynes -- this is easy to follow,
      // but we should revisit it and possibly reuse
      // the growth and theta checks that we made for the gts case.

      if (gte > 1.)
	{
	  coeff_e = -k*(k-1.);
	  if (growth > 1.)
	    arg1_e /= growth*growth;
	  else
	    coeff_e /= growth*growth;
	  if (theta > 1.)
	    arg1_e /= theta;
	  else
	    coeff_e /= theta;
	}
      else if (gte < -1.)
	{
	  arg1_e *= -k*(k-1.);
	  if (growth < -1.)
	    coeff_e /= growth*growth;
	  else
	    arg1_e /= growth*growth;
	  if (theta > 1.)
	    coeff_e /= theta;
	  else
	    arg1_e /= theta;
	}
      else // no over/underflow in exp()
	coeff_e = -k*(k-1.)/(theta*growth*growth); // arg1_e already set


      starttime = (*tit).endtime;
      result +=  coeff_s * SafeProductWithExp(arg1_s, gts)
	       - coeff_e * SafeProductWithExp(arg1_e, gte);
    }

  return result;
}

// erynes 2004/05/20 -- This method currently returns the following:
//
// sum_over_coalescent_time_intervals_of_population_whichparam(
//    (k_i(k_i - 1)/(2*theta))*(
//                  tis^2 * (1 + g*tis)  -  tie^2 * (1 + g*tie)
//                             )
//                                                            )
// where
//   k_i is the number of lineages of this population in time interval i
//   g   is the current estimate of the growth parameter for this pop.
//   theta is the current estimate of the theta parameter for this pop.
//   tis is the timepoint at the beginning of time interval i
//   tie is the timepoint at the end of time interval i
//
// This non-public method is used to handle the case in which fabs(g)
// is tiny--that is, less than GROWTH_EPSILON.  It is derived from the
// formula used in DWait by expanding exp(g*t) in a Taylor series
// about g == 0, multiplying through in the original formula,
// and dropping all terms of order g^2 and above.
// Brief tests with Mathematica suggest that the original formula
// converges very nicely to this linear formula for normal values
// of the timepoints, the difference being less than one part in one
// billion, dropping down to zero difference for g == 0 (using
// L'Hopital's rule).
//
// Please also see the comments for DWait.
//
double
GrowPL::DlnWaitForTinyGrowth(const vector < double >&param,
			     const TreeSummary * treedata, 
			     const long &whichparam)
{
  const double growth = param[whichparam];
  assert(fabs(growth) < GROWTH_EPSILON); // otherwise use DWait

  const long whicht = m_thetastart + whichparam - m_start;
  const double theta = param[whicht];
  assert(theta); // we divide by it below
  double k, te;
  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list <Interval> :: const_iterator tit = treesum.begin(); 

  double result = 0.0;
  double ts = 0.0;

  for( ; tit != treesum.end(); ++tit)
    {
      te = (*tit).endtime;
      k = (*tit).xpartlines[whicht];

      result += (k*(k-1.)/(2.*theta)) *
	(ts*ts*(1. + growth*ts) - te*te*(1. + growth*te));
      ts = te;
    }

  return result;
}




// This method presently returns:
//
// sum_over_time_intervals_of_this_population's_coalescences(endtime)
//
// This is the derivative of log(Point) with respect to growth.
//
double
GrowPL::DlnPoint (const vector < double >&param,
		  const TreeSummary * treedata, const long &whichparam)
{
  const Interval * treesum = treedata->GetCoalSummary()->GetLongPoint();
  const Interval * tit;
  double result = 0. ;
  // indicator to Theta, we need this to get the correct
  // time intervals in treesum
  long whicht = m_thetastart + whichparam - m_start;

  for(tit= treesum; tit != NULL; tit = tit->next)
    {
      if(whicht==(*tit).oldstatus)
	result += (*tit).endtime;
    }

  return result;
}


//-----------------------------------------------------
// RecombinePL: recombination forces specific waiting time and 
//              point probabilites and its derivatives
double
RecombinePL::lnWait (const vector < double >&param,
		     const TreeSummary * treedata)
{
  const vector<double>& rWait = treedata->GetRecSummary()->GetShortWait();
  return  -rWait[0] * *(param.begin () + m_start);
}

double
RecombinePL::lnPoint (const vector < double >&param, 
		      const vector < double >&lparam,
		      const TreeSummary * treedata)
{
  //  Sum_j(recevents * log(r))
  const vector<double>& nrec = treedata->GetRecSummary()->GetShortPoint();
  return nrec[0] * *(lparam.begin () + m_start);
}

double
RecombinePL::DlnWait (const vector < double >&param,
		      const TreeSummary * treedata, 
		      const long &whichparam)
{
  const vector<double>& rWait = treedata->GetRecSummary()->GetShortWait();
  return -rWait[0];
}

double
RecombinePL::DlnPoint (const vector < double >&param,
		       const TreeSummary * treedata, const long &whichparam)
{
  const vector<double>& nrecs = treedata->GetRecSummary()->GetShortPoint();
  if(nrecs[0]==0)
    {
      return 0.;
    }
  else
    {
      return SafeDivide(nrecs[0], *(param.begin() + m_start));
    }
}


//-----------------------------------------------------
// SelectPL: selection forces specific waiting time and 
//              point probabilites and its derivatives
double
SelectPL::lnWait (const vector < double >&param,
		  const TreeSummary * treedata)
{
  return 0.;
}

double
SelectPL::lnPoint (const vector < double >&param,
		   const vector < double >&lparam, 
		   const TreeSummary * treedata)
{
  return 0.;
}

double
SelectPL::DlnWait (const vector < double >&param,
		   const TreeSummary * treedata,
		 const long &whichparam)
{
  return 0.;
}

double
SelectPL::DlnPoint (const vector < double >&param,
		    const TreeSummary * treedata,
		    const long &whichparam)
{
  return 0.;
}

//------------------------------------------------------


double
CoalesceLogisticSelectionPL::lnWait(const vector<double>& param, 
			const TreeSummary *treedata)
{
  if (2 != m_nPop)
    {
      string msg = "CoalesceLogisticSelectionPL::lnWait() called with m_nPop = ";
      msg += ToString(m_nPop);
      msg += ".  m_nPop must equal 2, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator treesum_it;

  double result(0.), starttime(0.), s(param[m_s_is_here]);
  double theta_A0(param[m_start]), theta_a0(param[m_start+1]);
  double theta_total0(theta_A0 + theta_a0), dtau_A, dtau_a;

  if (fabs(s) < LOGISTIC_SELECTION_COEFFICIENT_EPSILON)
    return lnWaitForTinyLogisticSelectionCoefficient(param, treedata);

  for (treesum_it = treesum.begin(); treesum_it != treesum.end(); treesum_it++)
    {
      double endtime = treesum_it->endtime;
      const double& k_A = treesum_it->xpartlines[0]; // num. lineages, allele A
      const double& k_a = treesum_it->xpartlines[1]; // num. lineages, allele a

      dtau_A = (theta_A0*(endtime-starttime) + theta_a0*SafeExpDiff(s*endtime,s*starttime)/s)
	/theta_total0;
      dtau_a = (theta_a0*(endtime-starttime) - theta_A0*SafeExpDiff(-s*endtime,-s*starttime)/s)
	/theta_total0;

      result += -k_A*(k_A - 1.)*dtau_A/theta_A0 - k_a*(k_a - 1.)*dtau_a/theta_a0;

      starttime = endtime;
    }

  return result;
}

// This function expands lnWait() in a Taylor series about s = 0,
// so that we can smoothly handle the factor of 1/s in lnWait().
// This version retains only the first two terms in
// (exp(s*te) - exp(s*ts))/s = (te-ts) + (s/2!)(te^2 - ts^2) + (s^2/3!)(te^3 - ts^3) + ....
double
CoalesceLogisticSelectionPL::lnWaitForTinyLogisticSelectionCoefficient(const vector<double>& param,
								       const TreeSummary* treedata)
{
  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator treesum_it;

  double result(0.), starttime(0.), s(param[m_s_is_here]);
  double theta_A0(param[m_start]), theta_a0(param[m_start+1]), dtau_A, dtau_a;
  double firstOrderFactor_A = 0.5*theta_a0*s/(theta_A0 + theta_a0);
  double firstOrderFactor_a = -0.5*theta_A0*s/(theta_A0 + theta_a0);

  for (treesum_it = treesum.begin(); treesum_it != treesum.end(); treesum_it++)
    {
      double endtime = treesum_it->endtime;
      const double& k_A = treesum_it->xpartlines[0]; // num. lineages, allele A
      const double& k_a = treesum_it->xpartlines[1]; // num. lineages, allele a

      dtau_A = (endtime-starttime) + firstOrderFactor_A*(endtime*endtime - starttime*starttime);
      dtau_a = (endtime-starttime) + firstOrderFactor_a*(endtime*endtime - starttime*starttime);

      result += -k_A*(k_A - 1.)*dtau_A/theta_A0 - k_a*(k_a - 1.)*dtau_a/theta_a0;

      starttime = endtime;
    }

  return result;
}


double
CoalesceLogisticSelectionPL::lnPoint(const vector<double>& param,
				     const vector<double>& lparam,
				     const TreeSummary* treedata)
{
  if (2 != m_nPop)
    {
      string msg = "CoalesceLogisticSelectionPL::lnPoint() called with m_nPop = ";
      msg += ToString(m_nPop);
      msg += ".  m_nPop must equal 2, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  const Interval *treesum = treedata->GetCoalSummary()->GetLongPoint();
  const Interval *pTreesum;
  double theta_A0(param[0]), theta_a0(param[1]), logTheta_A0(lparam[0]),
    logTheta_a0(lparam[1]), logTheta_total0, s(param[m_s_is_here]);

  if (theta_A0 + theta_a0 > 0.)
    logTheta_total0 = log(theta_A0 + theta_a0);
  else
    {
      // This really, really should never ever happen.
      string msg = "CoalesceLogisticSelectionPL::lnPoint(), cannot take the ";
      msg += "natural logarithm of the sum theta_A0 + theta_a0, because theta_A0 = ";
      msg += ToString(theta_A0);
      msg += " and theta_a0 = " + ToString(theta_a0) + ".";
      throw impossible_error(msg);
    }

  double result = 0.0;

  for(pTreesum = treesum; pTreesum != NULL; pTreesum = pTreesum->next)
    {
      double thetaDenominator = theta_A0 +
	SafeProductWithExp(theta_a0, s * pTreesum->endtime);

      if (thetaDenominator > 0.)
	result += LOG2 - logTheta_total0 + log(thetaDenominator);
      else
	{
	  string msg = "CoalesceLogisticSelectionPL::lnPoint(), cannot take the ";
	  msg += "natural logarithm of the sum theta_A0 + theta_a0*exp(s*t), because theta_A0 = ";
	  msg += ToString(theta_A0) + ", theta_a0 = " + ToString(theta_a0) + ", s = ";
	  msg += ToString(s) + ", and t = " + ToString(pTreesum->endtime) + ".";
	  throw impossible_error(msg);
	}
      if (0L == pTreesum->oldstatus)
	result -= logTheta_A0;
      else if (1L == pTreesum->oldstatus)
	result -= logTheta_a0 + (pTreesum->endtime * s);
      else
	{
	  string msg = "CoalesceLogisticSelectionPL::lnPoint(), received a TreeSummary ";
	  msg += "containing an event with oldstatus = " + ToString(pTreesum->oldstatus);
	  msg += ".  \"oldstatus\" must be either 0 or 1, reflecting a coalescene in either ";
	  msg += "the population with allele A or the population with allele a.";
	  throw implementation_error(msg);
	}
    }

  return result;
}


double
CoalesceLogisticSelectionPL::DlnWait(const vector<double>& param,
				     const TreeSummary *treedata,
				     const long &whichparam)
{
  if (2 != m_nPop)
    {
      string msg = "CoalesceLogisticSelectionPL::DlnWait() called with m_nPop = ";
      msg += ToString(m_nPop);
      msg += ".  m_nPop must equal 2, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  double s(param[m_s_is_here]);

  if (fabs(s) < LOGISTIC_SELECTION_COEFFICIENT_EPSILON)
    return DlnWaitForTinyLogisticSelectionCoefficient(param, treedata, whichparam);

  bool derivativeWithRespectTo_A;

  if (0 == whichparam)
    derivativeWithRespectTo_A = true;
  else if (1 == whichparam)
    derivativeWithRespectTo_A = false;
  else
    {
      string msg = "CoalesceLogisticSelectionPL::DlnWait() called with whichparam = ";
      msg += ToString(whichparam);
      msg += ".  \"whichparam\" must equal 0 or 1, reflecting one population with the favored ";
      msg += "allele and one population with the disfavored allele.";
      throw implementation_error(msg);
    }

  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator treesum_it;
  double theta_A0(param[0]), theta_a0(param[1]), totalTheta_0(theta_A0+theta_a0);
  double dtau_A, dtau_a, endtime, starttime(0.), result(0.);

  for(treesum_it = treesum.begin(); treesum_it != treesum.end(); treesum_it++)
    {
      endtime = treesum_it->endtime;
      double dt = endtime - starttime;
      const double& k_A = treesum_it->xpartlines[0]; // num. lineages, allele A
      const double& k_a = treesum_it->xpartlines[1]; // num. lineages, allele a

      dtau_A = (theta_A0*(endtime-starttime) + theta_a0*SafeExpDiff(s*endtime,s*starttime)/s)
	/totalTheta_0;
      dtau_a = (theta_a0*(endtime-starttime) - theta_A0*SafeExpDiff(-s*endtime,-s*starttime)/s)
	/totalTheta_0;

      if (derivativeWithRespectTo_A)
	result += (k_A*(k_A - 1.)/theta_A0)*((dtau_A/theta_A0) - ((dt - dtau_A)/totalTheta_0))
	  + k_a*(k_a - 1.)*(dt - dtau_a)/(theta_A0*totalTheta_0);

      else
	result += k_A*(k_A - 1.)*(dt - dtau_A)/(theta_a0*totalTheta_0)
	  + (k_a*(k_a - 1.)/theta_a0)*((dtau_a/theta_a0) - (dt - dtau_a)/totalTheta_0);

      starttime = endtime;
    }

  return result;
}


double
CoalesceLogisticSelectionPL::DlnWaitForTinyLogisticSelectionCoefficient(const vector<double>& param,
									const TreeSummary *treedata,
									const long &whichparam)
{
  bool derivativeWithRespectTo_A;

  if (0 == whichparam)
    derivativeWithRespectTo_A = true;
  else if (1 == whichparam)
    derivativeWithRespectTo_A = false;
  else
    {
      string msg = "CoalesceLogisticSelectionPL::DlnWaitForTinyLogisticSelectionCoefficient() ";
      msg += "called with whichparam = ";
      msg += ToString(whichparam);
      msg += ".  \"whichparam\" must equal 0 or 1, reflecting one population with the favored ";
      msg += "allele and one population with the disfavored allele.";
      throw implementation_error(msg);
    }

  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator treesum_it;
  double theta_A0(param[0]), theta_a0(param[1]), totalTheta_0(theta_A0+theta_a0);
  double dtau_A, dtau_a, endtime, starttime(0.), result(0.), s(param[m_s_is_here]);
  double firstOrderFactor_A = 0.5*theta_a0*s/totalTheta_0;
  double firstOrderFactor_a = -0.5*theta_A0*s/totalTheta_0;

  for(treesum_it = treesum.begin(); treesum_it != treesum.end(); treesum_it++)
    {
      endtime = treesum_it->endtime;
      double dt = endtime - starttime;
      const double& k_A = treesum_it->xpartlines[0]; // num. lineages, allele A
      const double& k_a = treesum_it->xpartlines[1]; // num. lineages, allele a

      dtau_A = (endtime-starttime) + firstOrderFactor_A*(endtime*endtime - starttime*starttime);
      dtau_a = (endtime-starttime) + firstOrderFactor_a*(endtime*endtime - starttime*starttime);

      if (derivativeWithRespectTo_A)
	result += (k_A*(k_A - 1.)/theta_A0)*((dtau_A/theta_A0) - ((dt - dtau_A)/totalTheta_0))
	  + k_a*(k_a - 1.)*(dt - dtau_a)/(theta_A0*totalTheta_0);

      else
	result += k_A*(k_A - 1.)*(dt - dtau_A)/(theta_a0*totalTheta_0)
	  + (k_a*(k_a - 1.)/theta_a0)*((dtau_a/theta_a0) - (dt - dtau_a)/totalTheta_0);

      starttime = endtime;
    }

  return result;
}


double
CoalesceLogisticSelectionPL::DlnPoint(const vector<double>& param,
				      const TreeSummary *treedata,
				      const long &whichparam)
{
  if (2 != m_nPop)
    {
      string msg = "CoalesceLogisticSelectionPL::DlnPoint() called with m_nPop = ";
      msg += ToString(m_nPop);
      msg += ".  m_nPop must equal 2, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  bool derivativeWithRespectTo_A;

  if (0 == whichparam)
    derivativeWithRespectTo_A = true;
  else if (1 == whichparam)
    derivativeWithRespectTo_A = false;
  else
    {
      string msg = "CoalesceLogisticSelectionPL::DlnPoint() called with whichparam = ";
      msg += ToString(whichparam);
      msg += ".  \"whichparam\" must equal 0 or 1, reflecting one population with the major ";
      msg += "allele and one population with the minor allele.";
      throw implementation_error(msg);
    }

  const Interval *treesum = treedata->GetCoalSummary()->GetLongPoint();
  const Interval *pTreesum;
  double theta_A0(param[0]), theta_a0(param[1]), s(param[m_s_is_here]);
  double totalTheta_0(theta_A0+theta_a0), theta_A0squared(theta_A0*theta_A0);
  double result(0.);

  for (pTreesum = treesum; pTreesum != NULL; pTreesum = pTreesum->next)
    {
      double theta_a0_times_e_toThe_ste = SafeProductWithExp(theta_a0, s*pTreesum->endtime);
      double denominatorTerm = totalTheta_0*(theta_A0 + theta_a0_times_e_toThe_ste);

      if (derivativeWithRespectTo_A)
	{
	  if (0L == pTreesum->oldstatus) // coal. in allele A
	    result -= (theta_A0squared + (totalTheta_0 + theta_A0)*theta_a0_times_e_toThe_ste) /
	      (theta_A0*denominatorTerm);

	  else if (1L == pTreesum->oldstatus) // coal. in allele a
	    result -= (theta_a0_times_e_toThe_ste - theta_a0) / denominatorTerm;

	  else
	    {
	      string msg = "CoalesceLogisticSelectionPL::DlnPoint(), received a TreeSummary ";
	      msg += "containing an event with oldstatus = " + ToString(pTreesum->oldstatus);
	      msg += ".  \"oldstatus\" must be either 0 or 1, reflecting a coalescene in either ";
	      msg += "the population with allele A or the population with allele a.";
	      throw implementation_error(msg);
	    }
	}
      else
	{
	  if (0L == pTreesum->oldstatus) // coal. in allele A
	    result += theta_A0*(theta_a0_times_e_toThe_ste/theta_a0 - 1.) / denominatorTerm;

	  else if (1L == pTreesum->oldstatus) // coal. in allele a
	    result -= (theta_A0squared + theta_a0*(2.*theta_A0 + theta_a0_times_e_toThe_ste)) /
	      (theta_a0*denominatorTerm);

	  else
	    {
	      string msg = "CoalesceLogisticSelectionPL::DlnPoint(), received a TreeSummary ";
	      msg += "containing an event with oldstatus = " + ToString(pTreesum->oldstatus);
	      msg += ".  \"oldstatus\" must be either 0 or 1, reflecting a coalescene in either ";
	      msg += "the population with allele A or the population with allele a.";
	      throw implementation_error(msg);
	    }
	}
    }

  return result;
}

//------------------------------------------------------

double 
LogisticSelectionPL::lnWait(const vector<double>& param, 
			    const TreeSummary *treedata)
{
  return 0.;
}

double
LogisticSelectionPL::lnPoint(const vector<double>& param,
			     const vector<double>& lparam, 
			     const TreeSummary *treedata)
{
  return 0.;
}

double
LogisticSelectionPL::DlnWait(const vector<double>& param,
			     const TreeSummary *treedata,
			     const long &whichparam)
{
  if (whichparam != m_s_is_here)
    {
      string msg = "LogisticSelectionPL::DlnWait() called with whichparam = ";
      msg += ToString(whichparam);
      msg += ".  \"whichparam\" must equal " + ToString(m_s_is_here);
      msg += ", which corresponds to the selection coefficient \"s.\"";
      throw implementation_error(msg);
    }

  double s(param[m_s_is_here]);

  if (fabs(s) < LOGISTIC_SELECTION_COEFFICIENT_EPSILON)
    return DlnWaitForTinyLogisticSelectionCoefficient(param, treedata, whichparam);

  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator treesum_it;
  double theta_A0(param[0]), theta_a0(param[1]), denominator(s*s*(theta_A0+theta_a0));
  double ts(0.), te, result(0.);
  double term_s(1.), term_e;

  if (0. == denominator)
    {
      string msg = "LogisticSelectionPL::DlnWait(), attempted to divide by ";
      msg += "s*s*(theta_A0+theta_a0), where s = ";
      msg += ToString(s) + ", theta_A0 = ";
      msg += ToString(theta_A0) + " and theta_a0 = " + ToString(theta_a0) + ".";
      throw impossible_error(msg);
    }
  else if (0. == theta_A0 || 0. == theta_a0)
    {
      string msg = "LogisticSelectionPL::DlnWait(), theta_A0 = ";
      msg += ToString(theta_A0) + ", theta_a0 = " + ToString(theta_a0);
      msg += "; attempted to divide by zero.";
      throw impossible_error(msg);
    }

  for(treesum_it = treesum.begin(); treesum_it != treesum.end(); treesum_it++)
    {
      te = treesum_it->endtime;
      const double& k_A = treesum_it->xpartlines[0]; // num. lineages, allele A
      const double& k_a = treesum_it->xpartlines[1]; // num. lineages, allele a
      term_e = SafeProductWithExp(1. - s*te, s*te);

      result += (k_A*(k_A - 1.)/theta_A0)*theta_a0*(term_e - term_s);

      if (0. != term_e && 0. != term_s) // no underflow, so re-use these
	result -= (k_a*(k_a - 1.)/theta_a0)*theta_A0*( (1.-s*te)*(1.+s*te)/term_e
						       - (1.-s*ts)*(1.+s*ts)/term_s );
      else
	result -= (k_a*(k_a-1.)/theta_a0)*theta_A0*(SafeProductWithExp(1.+s*te,-s*te)
						    - SafeProductWithExp(1.+s*ts,-s*ts));
      ts = te;
      term_s = term_e;
    }

  return result/denominator;

}


double
LogisticSelectionPL::DlnWaitForTinyLogisticSelectionCoefficient(const vector<double>& param,
								const TreeSummary *treedata,
								const long &whichparam)
{
  const list<Interval>& treesum = treedata->GetCoalSummary()->GetLongWait();
  list<Interval>::const_iterator treesum_it;
  double theta_A0(param[0]), theta_a0(param[1]), theta_A0_over_theta_a0;
  double ts(0.), te, result(0.), s(param[m_s_is_here]);
  double ts_squared(0.), te_squared;

  if (0. == theta_A0 || 0. == theta_a0 || 0. == theta_A0 + theta_a0)
    {
      string msg = "LogisticSelectionPL::DlnWaitForTinySelectionCoefficient(), theta_A0 = ";
      msg += ToString(theta_A0) + ", theta_a0 = " + ToString(theta_a0);
      msg += "; attempted to divide by zero.";
      throw impossible_error(msg);
    }

  theta_A0_over_theta_a0 = theta_A0/theta_a0;

  for(treesum_it = treesum.begin(); treesum_it != treesum.end(); treesum_it++)
    {
      te = treesum_it->endtime;
      te_squared = te*te;
      const double& k_A = treesum_it->xpartlines[0]; // num. lineages, allele A
      const double& k_a = treesum_it->xpartlines[1]; // num. lineages, allele a

      result -= 0.5*(te_squared - ts_squared)*(k_A*(k_A-1.)/theta_A0_over_theta_a0 -
					       k_a*(k_a-1.)*theta_A0_over_theta_a0);
      result -= (s/3.)*(te_squared*te - ts_squared*ts)*(k_A*(k_A-1.)/theta_A0_over_theta_a0 +
							k_a*(k_a-1.)*theta_A0_over_theta_a0);
      ts= te;
      ts_squared = te_squared;
    }

  return result/(theta_A0 + theta_a0);
}


double
LogisticSelectionPL::DlnPoint(const vector<double>& param,
			      const TreeSummary *treedata,
			      const long &whichparam)
{
  if (whichparam != m_s_is_here)
    {
      string msg = "LogisticSelectionPL::DlnPoint() called with whichparam = ";
      msg += ToString(whichparam);
      msg += ".  \"whichparam\" must equal " + ToString(m_s_is_here);
      msg += ", which corresponds to the selection coefficient \"s.\"";
      throw implementation_error(msg);
    }

  const Interval *treesum = treedata->GetCoalSummary()->GetLongPoint();
  const Interval *pTreesum;
  double theta_A0(param[0]), theta_a0(param[1]), s(param[m_s_is_here]);
  double result(0.), term_a;

  for (pTreesum = treesum; pTreesum != NULL; pTreesum = pTreesum->next)
    {
      const double& te = pTreesum->endtime;
      term_a = SafeProductWithExp(theta_a0, s*te);

      if (0L == pTreesum->oldstatus) // coal. in allele A
	result += te*term_a/(theta_A0 + term_a); // rhs = 0 for s << 0

      else if (1L == pTreesum->oldstatus) // coal. in allele a
	result -= te/(term_a/theta_A0 + 1.);

      else
	{
	  string msg = "LogisticSelectionPL::DlnPoint(), received a TreeSummary ";
	  msg += "containing an event with oldstatus = " + ToString(pTreesum->oldstatus);
	  msg += ".  \"oldstatus\" must be either 0 or 1, reflecting a coalescence in either ";
	  msg += "the population with allele A or the population with allele a.";
	  throw implementation_error(msg);
	}
    }

  return result;
}

