// $Id: rangex.cpp,v 1.4 2007/04/27 19:59:52 lpsmith Exp $

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "rangex.h"
#include "stringx.h"
#include "registry.h"

using std::string;
using std::make_pair;

string ToString(rangepair rpair)
{
  //Note!  This prints stuff out in 'user units' instead of 'internal units'.
  // In other words, if the range is <0,34>, this is an 'open-ended' interval
  // and the actual active sites are 0:33
  //
  //Furthermore, if the user is expecting to have no zeroes in the output, we
  // need to convert the 0 to -1, and display "-1:33".
  rpair.second--;
  rpair = ToNoZeroesIfNeeded(rpair);
  string retval = ToString(rpair.first);
  if (rpair.second > rpair.first) {
    retval += ":" + ToString(rpair.second);
  }
  return retval;
}

string ToString(rangeset rset)
{
  rangeset::iterator rpair = rset.begin();
  if (rpair == rset.end()) {
    return "(none)";
  }
  string retval = ToString(*rpair);
  rpair++;
  for (; rpair != rset.end(); rpair++){
    retval += ", " + ToString(*rpair);
  }
  return retval;
}
  
rangeset AddPairToRange(rangepair addpart, rangeset rset)
{
  rangeset retset = rset;

  rangepair low  = make_pair(addpart.first, addpart.first);
  rangepair high = make_pair(addpart.second, addpart.second);

  long newlow = low.first;
  long newhigh = high.second;

  rangesetiter early = retset.lower_bound(low);
  if (early != retset.begin()) {
    early--;
    if (early->second >= low.first) {
      //'low' falls within early's interval
      newlow = early->first;
    }
  }
  
  rangesetiter late = retset.upper_bound(high);
  //We need to increment this late.first == high.first
  if (late != retset.end()) {
    if (late->first == high.first) {
      late++;
    }
  }
  if (late != retset.begin()) {
    late--;
    if (late->second > high.first) {
      //'high' falls within late's interval
      newhigh = late->second;
    }
  }

  early = retset.lower_bound(make_pair(newlow, newlow+1));
  late  = retset.upper_bound(make_pair(newhigh-1, newhigh));

  retset.erase(early, late);
  retset.insert(make_pair(newlow, newhigh));

  return retset;

}

rangeset RemovePairFromRange(rangepair removepart, rangeset rset)
{
  rangeset retset;
  for (rangeset::iterator range=rset.begin(); range != rset.end(); range++){
    if ((range->first >= removepart.first) &&
        (range->second <= removepart.second)) {
      //Don't add it.
    }
    else {
      if ((range->first < removepart.first) &&
          (range->second > removepart.second)) {
        //Add two outside halves.
        retset.insert(make_pair(range->first, removepart.first));
        retset.insert(make_pair(removepart.second, range->second));
      }
      else if ((range->second > removepart.first) &&
               (range->second <= removepart.second)) {
        //Add only the first half.
        retset.insert(make_pair(range->first, removepart.first));
      }
      else if ((range->first >= removepart.first) &&
               (range->first <= removepart.second)) {
        //Add only the second half.
        retset.insert(make_pair(removepart.second, range->second));
      }
      else {
        //Add the whole thing.
        retset.insert(*range);
      }
    }
  }
  return retset;
}

rangeset AddRangeToRange(rangeset addrange, rangeset rset)
{
  rangeset retset = rset;
  for (rangeset::iterator addpair = addrange.begin();
       addpair != addrange.end(); addpair++) {
    retset = AddPairToRange(*addpair, retset);
  }
  return retset;
}

rangeset RemoveRangeFromRange(rangeset removerange, rangeset rset)
{
  rangeset retset = rset;
  for (rangeset::iterator removepair = removerange.begin();
       removepair != removerange.end(); removepair++) {
    retset = RemovePairFromRange(*removepair, retset);
  }
  return retset;
}

long CountSites(rangeset rset)
{
  long retval(0);
  for (rangeset::iterator rpair = rset.begin(); rpair != rset.end(); rpair++) {
    retval += rpair->second - rpair->first;
  }
  return retval;
}


//These functions are to be used when converting site labels for display to
// the user, and from *menu* input from the user.  It is assumed that in the
// XML, the 'user input' has already been converted to the 'sequential' version.
long ToSequentialIfNeeded(long input)
{
  //Note:  input might equal zero if it was the upper end of a rangepair,
  // since rangepairs are open-ended by default.  In other words, if the user
  // types in "-20:-1" in the menu, the first thing that happens is that these
  // numbers are converted to the pair <-20, 0>.  To convert these values to
  // the 'sequential' numbering system, this needs to be converted to the
  // pair <-19, 1>.

  if (registry.GetConvertOutputToEliminateZeroes() && (input<=0)) {
    input++;
  }
  return input;
}

long ToNoZeroesIfNeeded(long input)
{
  if (registry.GetConvertOutputToEliminateZeroes() && (input <=0)) {
    input--;
  }
  return input;
}

rangepair ToSequentialIfNeeded(rangepair input)
{
  if (registry.GetConvertOutputToEliminateZeroes()) {
    input.first = ToSequentialIfNeeded(input.first);
    input.second = ToSequentialIfNeeded(input.second);
  }
  return input;
}

rangepair ToNoZeroesIfNeeded(rangepair input)
{
  if (registry.GetConvertOutputToEliminateZeroes()) {
    input.first = ToNoZeroesIfNeeded(input.first);
    input.second = ToNoZeroesIfNeeded(input.second);
  }
  return input;
}
