//$Id: range.cpp,v 1.14 2006/03/31 21:49:55 jay Exp $
/* 
   Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

   This software is distributed free of charge for non-commercial use
   and is copyrighted.  Of course, we do not guarantee that the software
   works, and are not responsible for any damage you may cause or have.
*/

#include "range.h"
#include "stringx.h"
#include "errhandling.h"

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

using namespace std;

//_______________________________________________________________
//_______________________________________________________________

Range::Range(long nsites, const vector<long>& alwaysact)
  : m_recsite(FLAGLONG),
    m_newlinks(0),
    m_side(unset_side),
    m_oldsitesset(false)
{
  rangepair allsites(0L, nsites);
  m_activesites.clear();
  m_activesites.insert(allsites);
  // set up the local-partition-force locations
  for (unsigned long i = 0; i < alwaysact.size(); ++i) {
    pair<long,long> p = make_pair(alwaysact[i],alwaysact[i]+1);
    m_alwaysactive.insert(p);
  }

} /* Range::Range(long) */

//_______________________________________________________________

void Range::ORAppend(rangepair newrange, rangeset& oldranges) const
{

  if (oldranges.empty()) {
    oldranges.insert(newrange);
    return;
  }

  rangeset::iterator last = --oldranges.end();

  assert(newrange.first >= last->first); // expect sequential,
  // sorted input

  // new is contained in old
  if (newrange.second <= last->second) return;

  // new is after old
  if (newrange.first > last->second) {
    oldranges.insert(oldranges.end(),newrange);
    return;
  }

  // new starts within old and extends past it
  newrange.first = last->first;
  oldranges.erase(last);
  oldranges.insert(oldranges.end(),newrange);
  return;

} /* Range::ORAppend */

//_______________________________________________________________

rangeset Range::OR(const rangeset& set1, const rangeset& set2) const
{

  if (set1.empty()) return set2;
  if (set2.empty()) return set1;

  rangeset mergeset;
  merge(set1.begin(),set1.end(),set2.begin(),set2.end(),
        inserter(mergeset,mergeset.begin()));

  rangeset newset;
  rangeset::iterator rit;
  for(rit = mergeset.begin(); rit != mergeset.end(); ++rit)
    ORAppend(*rit,newset);

  return newset;

} /* Range::OR */

//_______________________________________________________________

rangeset Range::AND(const rangeset& mother, const rangeset& father) const
{

  rangeset::const_iterator m = mother.begin();
  rangeset::const_iterator f = father.begin();

  rangeset result;
  rangepair newpair;

  if (mother.empty() || father.empty()) return result;

  while (true) {
    newpair.first = max((*m).first, (*f).first);
    newpair.second = min((*m).second, (*f).second);

    if (newpair.first < newpair.second)
      result.insert(result.end(),newpair);

    if ((*m).second < (*f).second) ++m;
    else ++f;

    if (m == mother.end() || f == father.end()) return result;
  }

} /* Range::AND */

//_______________________________________________________________

// Span returns the number of possible links in the range sent to it.
long Range::Span(const rangeset& sites) const
{
  if (sites.empty()) return 0L;

  long start = sites.begin()->first;
  long end = (--sites.end())->second;

  return end - start - 1;

} /* Range::Span */

//_______________________________________________________
// At the end of this function, Range::newactives
// should contain the sites that span newly active links,
// and Range::m_newlinks should contain a count of those links.

void Range::UpdateNewLinks(long nsites)
{
  assert(m_oldsitesset);
  
  m_newactivesites.clear();
  m_newlinks = 0;
  //We need to use both the active and the alwaysactive sites because
  // the alwaysactive sites always affect recombination eligibility.
  rangeset allactivesites = OR(m_activesites, m_alwaysactive);
  rangeset alloldactivesites = OR(m_oldsites, m_alwaysactive);

  // If there are no active sites, there are no newly active links
  if(allactivesites.empty()) return;

  // Calculate the sites spanning the old active links and
  // the sites spanning the current active links

  long newstart = allactivesites.begin()->first;
  long newend = (--allactivesites.end())->second;
  long oldstart = alloldactivesites.begin()->first;
  long oldend = (--alloldactivesites.end())->second;

  // if all links used to be active, none can be newly active, OR
  // if there is only 1 active site, there are no active links
  if ((oldstart == 0 && oldend == nsites) || (newstart == newend-1))
    return;

  // Find the sites spanning the old *inactive* links
  rangeset deadlinks;
  if (oldstart > 0) deadlinks.insert(rangepair(0,oldstart));
  if (oldend < nsites) deadlinks.insert(rangepair(oldend-1,nsites-1));

  assert(!deadlinks.empty());

  rangeset livelinks;
  livelinks.insert(rangepair(newstart,newend-1));

  // Calculate the intersection of old dead links and current
  // live links; these are newly active links
  m_newactivesites = AND(deadlinks, livelinks);

  // Compute the number of links involved
  rangeset::iterator i = m_newactivesites.begin();
  for ( ; i != m_newactivesites.end(); ++i) {
    m_newlinks += i->second - i->first;
  }

} /* Range::UpdateNewLinks */

//_________________________________________________

void Range::ClearOldsites()
{
  m_oldsites.clear();
  m_oldsitesset = false;

} /* Range::ClearOldsites */

//_________________________________________________

void Range::ClearNewactivesites()
{
  m_newactivesites.clear();
  m_newlinks = 0L;

} /* Range::ClearNewactivesites */

//_________________________________________________
//_________________________________________________

//The following routines only involve activesites and not alwaysactivesites.

void Range::SetOldsites(const rangeset& src)
{
  m_oldsites = src;
  m_oldsitesset = true;

}; /* SetOldsites */

//_________________________________________________

void Range::SetOldsites(const Range& src, long nsites)
{
  // As an optimization, src.m_oldsites may not be set if it
  // has never been relevant; if so our old active sites
  // are simply src's active sites.

  if (!src.m_oldsitesset) {
    m_oldsites = src.m_activesites;
  }
  else {
    m_oldsites = src.m_oldsites;
  }
  m_oldsitesset = true;

  UpdateNewLinks(nsites);

} /* Range::SetOldsites */

//_________________________________________________
bool Range::IsSiteActive(long site) const
{
  if (m_activesites.empty()) return false;

  rangeset::const_iterator range;
  for(range = m_activesites.begin(); range != m_activesites.end(); ++range) {
    if (range->second <= site) continue;

    if (range->first <= site) return true;
    break;
  }

  return false;

} /* Range::IsSiteActive */

//_________________________________________________
void Range::SetCRange(const Range& child1, const Range& child2)
{
  m_activesites = OR(child1.m_activesites,child2.m_activesites);

} /* Range::SetCRange */

//_________________________________________________

void Range::SetMRange(const Range& child1)
{
  m_activesites = child1.m_activesites;

} /* Range::SetMRange */

//_________________________________________________

void Range::SetRRange(const Range& child1, const rangepair& newsites,
                      bool isleft)
{
  rangeset newset;
  newset.insert(newsites);
  m_activesites = AND(child1.m_activesites,newset);
  if (isleft) {
    m_side = left_side;
    m_recsite = newsites.second - 1;
  }
  else {
    m_side = right_side;
    m_recsite = newsites.first - 1;
  }
} /* Range::SetRRange */

//_________________________________________________

bool Range::UpdateCRange(const Range& child1, const Range& child2,
                         long nsites)
{
  rangeset newsites = OR(child1.m_activesites,child2.m_activesites);

  if (newsites == m_activesites) return false;

  if (!m_oldsitesset) {
    m_oldsites = m_activesites;
    m_oldsitesset = true;
  }

  m_activesites = newsites;

  UpdateNewLinks(nsites);

  return true;

} /* Range::UpdateCRange */

//_________________________________________________

bool Range::UpdateMRange(const Range& child1, long nsites)
{
  if (m_activesites == child1.m_activesites) return false;

  if (!m_oldsitesset) {
    m_oldsites = m_activesites;
    m_oldsitesset = true;
  }

  m_activesites = child1.m_activesites;

  UpdateNewLinks(nsites);

  return true;

} /* Range::UpdateMRange */

//_________________________________________________

bool Range::UpdateRRange(const Range& child1, long nsites)
{
  if (!m_oldsitesset) {
    m_oldsites = m_activesites;
    m_oldsitesset = true;
  }

  rangeset recrange;
  if (m_side == left_side)
    recrange.insert(rangepair(0L,m_recsite+1));
  else if (m_side == right_side)
    recrange.insert(rangepair(m_recsite+1,nsites));
  else {
    assert(false); //A recombination without a set side?
    throw implementation_error("A recombination was specified without knowledge of which side of the recombination we were on.");
  }

  m_activesites = AND(child1.m_activesites,recrange);

  UpdateNewLinks(nsites);

  return true;

} /* Range::UpdateRRange */

//_________________________________________________

bool Range::operator==(const Range& other) const
{
  return (m_activesites == other.m_activesites &&
          m_recsite == other.m_recsite);
}

//_________________________________________________
//_________________________________________________

//The following routines use both activesites and alwaysactivesites.

long Range::GetFirstActivesite() const
{
  if (m_alwaysactive.empty()) {
    assert(!m_activesites.empty());
    return (m_activesites.begin()->first);
  }
  if (m_activesites.empty()) {
    assert(!m_alwaysactive.empty());
    return (m_alwaysactive.begin()->first);
  }
  return std::min( m_activesites.begin()->first,
                   m_alwaysactive.begin()->first);
}

//_________________________________________________
bool Range::NoActivesites() const
{
  if (!(m_activesites.empty())) {
    return false;
  }
  for (set<rangepair>::const_iterator ri = m_alwaysactive.begin();
       ri != m_alwaysactive.end(); ri++) {
    switch(m_side) {
    case right_side:
      if (m_recsite < (*ri).first) {
        return false;
      }
      break;
    case left_side:
      if ((*ri).first <= m_recsite) {
        return false;
      }
      break;
    case unset_side:
      //For a non-recombination branch, any always-active sites means
      // we always return false from this routine.
      return false;
    }
  }
  //cout << endl << "No active sites:  ";
  //PrintInfo();
  return true;
};

//_________________________________________________

bool Range::IsAlwaysActiveInRange() const
{
  for (set<rangepair>::const_iterator ri = m_alwaysactive.begin();
       ri != m_alwaysactive.end(); ri++) {
    switch(m_side) {
    case right_side:
      if (m_recsite < (*ri).first) { //Warning!  Won't work with wide ranges!
        return true;
      }
      break;
    case left_side:
      if ((*ri).first <= m_recsite) {
        return true;
      }
      break;
    case unset_side:
      //For a non-recombination branch, any always-active sites means
      // we always return true from this routine.
      return true;
    }
  }
  return false;
}
  

long Range::ActiveLinks() const
{
  return Span(OR(m_activesites, m_alwaysactive));
}

//_________________________________________________

void Range::PrintActives() const
{
  cout << "Active sites:  " << ToString(m_activesites) << endl;
} /* Range::PrintActives */

//_________________________________________________

void Range::PrintNewActives() const
{
  cout << "New active sites:  " << ToString(m_newactivesites) << endl;

} /* Range::PrintNewActives */

//_________________________________________________

void Range::PrintAlwaysActives() const
{
  cout << "Always active sites:  " << ToString(m_alwaysactive) << endl;

} /* Range::PrintNewActives */

//_________________________________________________

void Range::PrintRightOrLeft() const
{
  cout << "Right or left:  ";
  switch(m_side) {
  case right_side:
    cout << "right";
    break;
  case left_side:
    cout << "left";
    break;
  case unset_side:
    cout << "unset";
    break;
  }
  cout << endl;
}
//_________________________________________________

void Range::PrintInfo() const
{
  PrintActives();
  PrintAlwaysActives();
  cout << "Rec site = " << m_recsite << endl;
  PrintRightOrLeft();
}
