//$Id: rectree.cpp,v 1.45 2007/04/23 18:47:40 jay Exp $

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "errhandling.h"  // can throw overrun_error
#include "force.h"        // for TimeSize object
#include "range.h"
#include "runreport.h"
#include "tree.h"
#include "mathx.h"
#include "branchbuffer.h" // used in SetActiveLinksFrom for
                          //    BranchBuffer::ExtractConstBranches

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

typedef boost::shared_ptr<RBranch> RBranch_ptr;

using namespace std;

//___________________________________________________________________________

RecTree::RecTree()
  : Tree(),
    m_pMovingLocusVec(),
    m_protoMovingCells()
{
  activeLinks = 0;
  openLinks   = 0;
}

//____________________________________________________________________________

RecTree::RecTree(const RecTree &tree, bool makestump)
  : Tree(tree, makestump),
    m_pMovingLocusVec(tree.m_pMovingLocusVec),
    m_protoMovingCells(tree.m_protoMovingCells)

{
  activeLinks = 0;
  openLinks   = 0;
}

//____________________________________________________________________________

Tree *RecTree::Clone() const
{
  RecTree *tree = new RecTree(*this,false);
  return tree;
}

//____________________________________________________________________________

Tree *RecTree::MakeStump() const
{
  RecTree *tree = new RecTree(*this,true);
  return tree;
}

//____________________________________________________________________________

void RecTree::Clear()
{
  Tree::Clear();
  activeLinks = 0;
  openLinks   = 0;
}

//____________________________________________________________________________

void RecTree::CopyTips(const Tree *tree)
{
  Tree::CopyTips(tree);
  activeLinks = 0;
  openLinks   = 0;
  for(vector<Locus>::const_iterator locus = m_pMovingLocusVec->begin();
      locus != m_pMovingLocusVec->end(); ++locus) {
    m_aliases.push_back(locus->GetDLCalc()->RecalculateAliases(*this, *locus));
  }
}

//____________________________________________________________________________

void RecTree::CopyBody(const Tree *tree)
{
  Tree::CopyBody(tree);
  activeLinks = 0;
  openLinks   = 0;
}

//____________________________________________________________________________

void RecTree::CopyPartialBody(const Tree *tree)
{
  Tree::CopyPartialBody(tree);
  activeLinks = 0;
  openLinks   = 0;
}

//____________________________________________________________________________

void RecTree::RecurseBreak(Branch_ptr pBranch)
{
  Branch_ptr pParent = pBranch->Parent(0);

  if (pParent->CanRemove(pBranch))
  {
    RecurseBreak(pParent);                                  // Recursion
    m_timelist.Remove(pParent);

    pParent = pBranch->Parent(1);
    if (pParent)
    {
      RecurseBreak(pParent);                                // Recursion
      m_timelist.Remove(pParent);
    }
  }

}

//____________________________________________________________________________

void RecTree::Break(Branch_ptr pBranch)
{
  Branch_ptr pParent = pBranch->Parent(0);

  // now update the activesite and newlyactivesite information
  if (pParent->Event() != btypeCoal)
    pParent->UpdateRange(m_totalsites);
  else {
    assert(boost::dynamic_pointer_cast<CBranch>(pParent));
    Branch_ptr br = boost::dynamic_pointer_cast<CBranch>(pParent)->
       OtherChild(pBranch);
    pParent->m_range.UpdateMRange(br->m_range,m_totalsites);
  }
  UpdateRange(pParent);

  if (pParent->CanRemove(pBranch))
  {
    RecurseBreak(pParent);
    m_timelist.Remove(pParent);

    pParent = pBranch->Parent(1);
    if (pParent)
    {
      RecurseBreak(pParent);
      m_timelist.Remove(pParent);
    }
  }

}

//____________________________________________________________________________

const vector<LocusCell>& RecTree::CollectMovingCells()
{
  if (m_protoMovingCells.empty()) {
    unsigned long i;
    for (i = 0; i < m_pMovingLocusVec->size(); ++i) {
      m_protoMovingCells.push_back((*m_pMovingLocusVec)[i].GetProtoCell());
    }
  }
  return m_protoMovingCells;

} /* CollectCells */

//____________________________________________________________________________

vector<Branch_ptr> RecTree::ActivateTips(Tree* othertree)
{
  vector<Branch_ptr> tips = Tree::ActivateTips(othertree);
  activeLinks = m_timelist.GetNTips() * (m_totalsites - 1);
  // -1 because there is no link after last site
  openLinks = 0;
  return tips;
}

//____________________________________________________________________________

Branch_ptr RecTree::ActivateBranch(Tree* othertree)
{
  Branch_ptr pActive = Tree::ActivateBranch(othertree);
  activeLinks += pActive->m_range.ActiveLinks();
  return pActive;
}

//____________________________________________________________________________

Branch_ptr RecTree::ActivateRoot()
{
  Branch_ptr pRoot = Tree::ActivateRoot();
  activeLinks += pRoot->m_range.ActiveLinks();
  openLinks = 0;
  return pRoot;
}

//____________________________________________________________________________

void RecTree::AttachBase(Branch_ptr newroot)
{
  Tree::AttachBase(newroot);
  activeLinks = 0;
}

//____________________________________________________________________________

void RecTree::UpdateRange(Branch_ptr pBranch)
{
  Branch_ptr pParent = pBranch->Parent(0);
  if (pParent == m_timelist.Base())
    return;

  if (pParent->UpdateRange(m_totalsites))
    UpdateRange(pParent);

  pParent = pBranch->Parent(1);
  if (pParent)
    if (pParent->UpdateRange(m_totalsites))
      UpdateRange(pParent);
}

//____________________________________________________________________________

vector<Branch_ptr> RecTree::FirstInterval(double eventT)
{
  Branch_ptr pBranch;

  vector<Branch_ptr> newinactives = Tree::FirstInterval(eventT);

  // Initialize the count of open links.
  openLinks = 0;

  // Update the the site ranges in the lineage below the break.
  Branchiter brit;
  for (brit = m_timelist.FirstBody(); brit != m_timelist.End();
    brit = m_timelist.NextBody(brit))
  {
    pBranch = *brit;
    if (pBranch->m_marked)
    {
      if (pBranch->UpdateRange(m_totalsites))
        UpdateRange(pBranch);
    }
  }
  return newinactives;
}

//____________________________________________________________________________

void RecTree::NextInterval(Branch_ptr pBranch)
{
  assert(pBranch->Event() != btypeStick); 

  Branch_ptr pParent = pBranch->Parent(0);

  openLinks -= pParent->Child(0)->m_range.NewActiveLinks();
  openLinks += pParent->m_range.NewActiveLinks();

  // if the branch has a sibling, remove it
  if (pParent->Child(1))       
  {
    openLinks -= pParent->Child(1)->m_range.NewActiveLinks();
  }

  // if the branch has a second parent, insert it
  else if (pBranch->Parent(1))
  {
    openLinks += pBranch->Parent(1)->m_range.NewActiveLinks();
  }

}

//____________________________________________________________________________

void RecTree::Prune()
{
  Branch_ptr pBranch, pChild;

  openLinks = 0;

  Branchiter brit;
  for (brit = m_timelist.FirstBody(); brit != m_timelist.End(); )
  {
    // Remove branches that have no active sites
    pBranch = *brit;
    // clean up old range information
    pBranch->m_range.ClearOldsites();
    pBranch->m_range.ClearNewactivesites();

    if (pBranch->m_range.NoActivesites())
    {
      pChild = pBranch->Child(0);
      if (pChild->Parent(0) == pBranch)
        pChild->SetParent(0,pChild->Parent(1));
      pChild->SetParent(1,Branch::NONBRANCH);
      pChild->Parent(0)->m_marked = true;
      RecurseBreak(pBranch);     // must be called in the order
      brit = m_timelist.NextBody(brit);// RecurseBreak() then iterator set
      m_timelist.Remove(pBranch);  // then list removal since RecurseBreak()
    }                            // may invalidate iterators

    else
    {
      brit = m_timelist.NextBody(brit);
    }
  }

  Tree::Prune();

}

void RecTree::ReassignDLsFor(string lname, long marker, long ind)
{
  //Find out which locus we're dealing with.  We have its name.
  // If the locus in question is in m_pLocusVec, we call through to
  // Tree::ReassignDLsFor
  long locus = FLAGLONG;
  for (unsigned long lnum=0; lnum<m_pMovingLocusVec->size(); lnum++) {
    if ((*m_pMovingLocusVec)[lnum].GetName() == lname) {
      locus = lnum;
    }
  }
  if (locus == FLAGLONG) {
    //the locus must be in m_pLocusVec--call the base function
    return Tree::ReassignDLsFor(lname, marker, ind);
  }

  vector<Branch_ptr> haps = m_individuals[ind].GetAllTips();
  vector<LocusCell> cells = m_individuals[ind].GetLocusCellsFor(lname, marker);

  for (unsigned long tip=0; tip<haps.size(); tip++) {
    Cell_ptr origcell = haps[tip]->GetDLCell(locus, 0, true);
    Cell_ptr newcell  = cells[tip][0];
    origcell->SetSiteDLs(marker, newcell->GetSiteDLs(marker));
    MarkForDLRecalc(haps[tip]);
  }

  // Don't recalculate aliases, since there are no aliases for the moving locus.
} /* RecTree::ReassignDLsFor */

//____________________________________________________________________________

Branch_ptr RecTree::CoalesceActive(double eventT, long maxEvents,
   Branch_ptr active1, Branch_ptr active2)
{
  Branch_ptr pBranch = Tree::CoalesceActive(eventT,maxEvents,active1,active2);

  pBranch->SetMovingDLCells(CollectMovingCells());

  pBranch->m_range.SetCRange(pBranch->Child(0)->m_range,
                           pBranch->Child(1)->m_range);

  activeLinks -= pBranch->Child(0)->m_range.ActiveLinks();
  activeLinks -= pBranch->Child(1)->m_range.ActiveLinks();
  activeLinks += pBranch->m_range.ActiveLinks();
  return pBranch;
}

//____________________________________________________________________________

Branch_ptr RecTree::CoalesceInactive(double eventT, long maxEvents,
   Branch_ptr active, Branch_ptr inactive)
{
  Branch_ptr pBranch = Tree::CoalesceInactive(eventT,maxEvents,active,inactive);

  pBranch->SetMovingDLCells(CollectMovingCells());

  pBranch->m_range.SetCRange(pBranch->Child(0)->m_range,
                           pBranch->Child(1)->m_range);

  pBranch->m_range.SetOldsites(pBranch->Child(0)->m_range, m_totalsites);

  if (!pBranch->HasSameActive(*(pBranch->Child(0))))
  {
    UpdateRange(pBranch);
  }

  openLinks   -= pBranch->Child(0)->m_range.NewActiveLinks();
  activeLinks -= pBranch->Child(1)->m_range.ActiveLinks();
  openLinks   += pBranch->m_range.NewActiveLinks();
  return pBranch;
}

//____________________________________________________________________________

Branch_ptr RecTree::Migrate(double eventT, long topop, long maxEvents,
   Branch_ptr active)
{
  Branch_ptr pBranch = Tree::Migrate(eventT, topop, maxEvents, active);

  pBranch->m_range.SetMRange(pBranch->Child(0)->m_range);

  // a migration does not change the active or newly active links
  return pBranch;
}

//____________________________________________________________________________

Branch_ptr RecTree::DiseaseMutate(double eventT, long endstatus,
   long maxEvents, Branch_ptr active)
{
  Branch_ptr pBranch = Tree::DiseaseMutate(eventT,endstatus,maxEvents,active);

  pBranch->m_range.SetMRange(pBranch->Child(0)->m_range);

  // a disease mutation does not change the active or newly active links

  return pBranch;
}

//____________________________________________________________________________

branchpair RecTree::RecombineActive(double eventT, long maxEvents, 
   FPartMap fparts, Branch_ptr pActive, long rec)
{
  if (NumberOfRecombinations() >= maxEvents) {
    rec_overrun e;
    throw e;
  }

  // Create the left parent of the active branch.  ---------------
  // saving partition handling til end

  RBranch_ptr pParentA = RBranch_ptr(new RBranch(pActive,rec,true,
     m_totalsites));
  pParentA->m_eventtime  = eventT;              // Set the event time.
  pParentA->SetChild(0,pActive);
  pActive->SetParent(0,pParentA);

  // Create the right parent of the active branch.  ---------------
  // saving partition handling til end
  RBranch_ptr pParentB = RBranch_ptr(new RBranch(pActive,rec,false,
     m_totalsites));
  pParentB->m_eventtime  = eventT;              // Set the event time.
  pParentB->SetChild(0,pActive);
  pActive->SetParent(1,pParentB);

  // now deal with partitions
  if (fparts.empty()) {
     pParentA->CopyPartitionsFrom(pActive);
     pParentB->CopyPartitionsFrom(pActive);
  } else {
     pParentA->RecCopyPartitionsFrom(pActive,fparts,true);
     pParentB->RecCopyPartitionsFrom(pActive,fparts,false);
  }

  activeLinks -= pActive->m_range.ActiveLinks();
  activeLinks += pParentA->m_range.ActiveLinks();
  activeLinks += pParentB->m_range.ActiveLinks();

  m_timelist.Collate(pParentA);
  m_timelist.Collate(pParentB);
  return make_pair(pParentA,pParentB);

}

//____________________________________________________________________________

branchpair RecTree::RecombineInactive(double eventT, long maxEvents,
   FPartMap fparts, Branch_ptr pBranch, long rec)
{
  if (NumberOfRecombinations() >= maxEvents) {
    rec_overrun e;
    throw e;
  }

  // This is the branch already in the tree
  bool parentA_is_left = m_randomSource->Bool();
  //All this might eventually be one constructor call.
  RBranch_ptr pParentA = RBranch_ptr(new RBranch(pBranch,rec,parentA_is_left,
     m_totalsites));
  pParentA->m_eventtime  = eventT;
  pParentA->CopyPartitionsFrom(pBranch);

  pParentA->SetChild(0,pBranch);
  pBranch->Parent(0)->ReplaceChild(pBranch, pParentA);
  pBranch->SetParent(0,pParentA);
  if (pBranch->Parent(1))
  {
    pBranch->Parent(1)->ReplaceChild(pBranch, pParentA);
    pBranch->SetParent(1,Branch::NONBRANCH);
  }

  pParentA->m_range.SetOldsites(pBranch->m_range, m_totalsites);
  UpdateRange(pParentA);

  // This is the new, active branch
  RBranch_ptr pParentB = RBranch_ptr(new RBranch(pBranch,rec,!parentA_is_left,
     m_totalsites));
  pParentB->m_eventtime  = eventT;
 
  pParentB->SetChild(0,pBranch);
  pBranch->SetParent(1,pParentB);

  if (fparts.empty()) {
    pParentB->CopyPartitionsFrom(pBranch);
  } else {
    pParentB->RecCopyPartitionsFrom(pBranch,fparts,!parentA_is_left);
  }

  // Now put both branches in place
  m_timelist.Collate(pParentA);
  m_timelist.Collate(pParentB);

  openLinks   -= pBranch->m_range.NewActiveLinks();
  openLinks   += pParentA->m_range.NewActiveLinks();
  activeLinks += pParentB->m_range.ActiveLinks();

  return make_pair(pParentA,pParentB);
}

//______________________________________________

rangevector RecTree::GetLocusSubtrees(rangepair span) const
{
  // Get recombination breakpoints
  set<long> breakpt(GetBreakpoints());

  // Push rangepairs representing each breakpoint into subtree vector.
  // The (pt + 1) avoids creating a rangepair starting at the last site.
  // The strange iterator manuver is because (pt + 1) is not allowed in g++.
  // (We do not risk an empty set; we know it contains at least 2 elements.)

  rangevector subtrees;
  set<long>::const_iterator pt = breakpt.begin();
  set<long>::const_iterator nextpt = pt;
  ++nextpt;
  long begin = span.first;
  long end = span.second;

  for ( ; nextpt != breakpt.end(); ++pt, ++nextpt) {
    // no overlap so we don't add it
    if (*nextpt <= begin || *pt >= end) continue; 
    // some overlap, but ends may need adjustment
    long first = *pt;
    long last = *nextpt;
    if (first < begin) first = begin;
    if (last > end) last = end;
    subtrees.push_back(rangepair(first, last));
  }
  return subtrees;

} /* GetLocusSubtrees */

void RecTree::SetMovingLocusVec(vector<Locus>* loc)
{
  m_pMovingLocusVec = loc;
  //We don't have to worry about the range--the moving loci move around within
  // the range of the fixed loci.
} /* SetMovingLocusVec */

void RecTree::SetMovingMapposition(long mloc, long site)
{
  assert(mloc < static_cast<long>(m_pMovingLocusVec->size()));
  (*m_pMovingLocusVec)[mloc].SetRegionalMapposition(site);
}
//______________________________________________

bool RecTree::DoesThisLocusJump(long mloc) const
{
  assert(mloc < static_cast<long>(m_pMovingLocusVec->size()));
  return ((*m_pMovingLocusVec)[mloc].GetAnalysisType() == mloc_mapjump);
}

//______________________________________________

bool RecTree::AnyRelativeHaplotypes() const
{
  for (IndVec::const_iterator ind = m_individuals.begin();
       ind != m_individuals.end(); ind++) {
    if (ind->MultipleTraitHaplotypes()) return true;
  }
  return false;
}

//______________________________________________

void RecTree::CalculateDataLikes()
{
  m_overallDL = CalculateDataLikesForFixedLoci();
  //The base function accumulates a likelihood into m_overallDL
  for (unsigned long loc = 0; loc<m_pMovingLocusVec->size(); ++loc) {
    if ((*m_pMovingLocusVec)[loc].GetAnalysisType() == mloc_mapjump) {
      m_overallDL += CalculateDataLikesForMovingLocus(loc);
    }
  }
  m_timelist.ClearUpdateDLs(); // reset the updating flags
}

double RecTree::CalculateDataLikesForMovingLocus(long loc)
{
  double likelihood;
  
  const Locus& locus = (*m_pMovingLocusVec)[loc];
  try {  // check for need to switch to normalization
    likelihood = locus.CalculateDataLikelihood(*this, true);
  }
  catch (datalikenorm_error& ex) { // normalization is set by thrower
    m_timelist.SetAllUpdateDLs();  // all subsequent loci will recalculate
                                   // the entire tree
    RunReport& runreport = registry.GetRunReport();
    runreport.ReportChat("\n", 0);
    runreport.ReportChat("Subtree of likelihood 0.0 found:  Turning on normalization and re-calculating.");

    likelihood = locus.CalculateDataLikelihood(*this, true);
  }
  return likelihood;

} /* CalculateDataLikesForMovingLocus */

//----------------------------------------------------------------
set<long> RecTree::GetBreakpoints() const
{
  // Create a (sorted) set containing all recombination breakpoints
  
  set<long> breakpt;
  breakpt.insert(0);
  breakpt.insert(m_totalsites);

  Branchconstiter brit;
  long site;
  for (brit = m_timelist.FirstBody(); brit != m_timelist.End();
    brit = m_timelist.NextBody(brit))
  {
    site = (*brit)->GetRecSite();
    if (site != FLAGLONG)    // FLAGLONG means there is no recombination here
      breakpt.insert(site+1);// inserting site+1 because of open interval 
  }                          // conventions

  return breakpt;
} /* GetBreakpoints */

DoubleVec2d RecTree::GetMapSummary()
{
  DoubleVec2d retvec;
  DoubleVec1d zeroes(m_totalsites, 0.0);
  for (unsigned long mloc=0; mloc<m_pMovingLocusVec->size(); mloc++) {
    DoubleVec1d mlikes = zeroes;
    long currentsite;
    switch((*m_pMovingLocusVec)[mloc].GetAnalysisType()) {
    case mloc_mapjump:
      //Find out where it is, make the likelihood at that value 1..
      currentsite = (*m_pMovingLocusVec)[mloc].GetRegionalMapposition();
      mlikes[currentsite] = 1.0;
      break;
    case mloc_mapfloat:
      //Calculate the likelihoods over all subtrees, return vectors with it.
      //mlikes = CalculateDataLikesWithRandomHaplotypesForFloatingLocus(mloc);
      CalculateDataLikesForAllHaplotypesForFloatingLocus(mloc, mlikes);
      break;
    case mloc_data:
    case mloc_partition:
      assert(false);
      throw implementation_error("We seem to want to collect mapping data for a segment without that type of analysis.  This is our fault; e-mail us at lamarc@gs.washington.edu");
    }
    retvec.push_back(mlikes);
  }
  return retvec;
}

DoubleVec1d RecTree::CalculateDataLikesWithRandomHaplotypesForFloatingLocus(long mloc)
{
  RandomizeMovingHaplotypes(mloc);
  //LS DEBUG
  //PrintTipData(mloc, 0);
  return CalculateDataLikesForFloatingLocus(mloc);
}

void RecTree::CalculateDataLikesForAllHaplotypesForFloatingLocus(long mloc, DoubleVec1d& mlikes)
{
  long ind = 0;
  //We need to change the vector of mlikes from zeroes to EXPMINS.
  mlikes = SafeLog(mlikes);
  UpdateDataLikesForIndividualsFrom(ind, mloc, mlikes);
  //LS DEBUG
  //PrintTipData(mloc,0);
  //LS TEST
  //cout << "Sum:  " << endl << ToString(mlikes, 5) << endl;
}

bool RecTree::UpdateDataLikesForIndividualsFrom(long ind, long mloc,
                                                DoubleVec1d& mlikes)
{
  if (static_cast<unsigned long>(ind) == m_individuals.size()) return true;
  string lname = (*m_pMovingLocusVec)[mloc].GetName();
  //LS NOTE: If this assert fires, we are mapping something with more
  // than one marker.  Which might be OK, but needs to be checked.
  assert ((*m_pMovingLocusVec)[mloc].GetNmarkers() == 1);
  for (long marker = 0; marker<(*m_pMovingLocusVec)[mloc].GetNmarkers();
       marker++) {
    bool newhaps = true;
    for (m_individuals[ind].ChooseFirstHaplotypeFor(lname, marker);
         newhaps;
         newhaps = m_individuals[ind].ChooseNextHaplotypeFor(lname, marker)) {
      ReassignDLsFor(lname, marker, ind);
      if (UpdateDataLikesForIndividualsFrom(ind+1, mloc, mlikes)) {
        //We're on the last one--update the data likelihood
        DoubleVec1d newlikes = CalculateDataLikesForFloatingLocus(mloc);
        //LS TEST:
        //cout << ToString(newlikes, 5) << endl;
        mlikes = AddValsOfLogs(mlikes, newlikes);
      }
    }
  }
  return false;
}


DoubleVec1d RecTree::CalculateDataLikesForFloatingLocus(long mloc)
{
  m_timelist.SetAllUpdateDLs();
  // We always need these on because there's only enough storage for a
  // single subtree, and the last subtree we used was at the other end
  // of the range.
  DoubleVec1d datalikes(m_totalsites);
  rangeset allowedranges = (*m_pMovingLocusVec)[mloc].GetAllowedRange();
  set<long> breakpoints = GetBreakpoints();
  breakpoints = IgnoreDisallowedSubTrees(breakpoints, allowedranges);
  set<long>::iterator left(breakpoints.begin()), right(breakpoints.begin());
  right++;
  for (;right != breakpoints.end(); left++, right++) {
    (*m_pMovingLocusVec)[mloc].SetRegionalMapposition(*left);
    double datalike = CalculateDataLikesForMovingLocus(mloc);
    for (long siteindex = *left; siteindex<*right; siteindex++) {
      datalikes[siteindex] = datalike;
    }
  }
  datalikes = ZeroDisallowedSites(datalikes, allowedranges);

  m_timelist.ClearUpdateDLs();
  //LS NOTE: We could theoretically not call ClearUpdateDLs if there was only
  // one subtree, but only if the *next* call to this function also had the
  // exact same subtree, which we can't enforce.  So, always call ClearUpdateDLs
  return datalikes;
}

set<long> RecTree::IgnoreDisallowedSubTrees(set<long> breakpoints,
                                            rangeset allowedranges)
{
  set<long> newbreakpoints;
  newbreakpoints.insert(allowedranges.begin()->first);
  newbreakpoints.insert(m_totalsites); //to close out the final range.
  for (rangeset::iterator range=allowedranges.begin();
       range != allowedranges.end(); range++) {
    for (long site = range->first; site<range->second; site++) {
      if (breakpoints.find(site) != breakpoints.end()) {
        newbreakpoints.insert(site);
      }
    }
  }
  return newbreakpoints;
}

DoubleVec1d RecTree::ZeroDisallowedSites(DoubleVec1d datalikes,
                                         rangeset allowedranges)
{
  DoubleVec1d newdatalikes(datalikes.size(), -DBL_BIG);
  for (rangeset::iterator range = allowedranges.begin();
       range != allowedranges.end(); range++) {
    for (long site = range->first; site < range->second; site++) {
      newdatalikes[site] = datalikes[site];
    }
  }
  return newdatalikes;
}

void RecTree::RandomizeMovingHaplotypes(long mlocus)
{
  //This function loops over all individuals in m_individuals and randomizes
  // the haplotypes.
  string lname = (*m_pMovingLocusVec)[mlocus].GetName();
  long nmarkers = (*m_pMovingLocusVec)[mlocus].GetNmarkers();
  for (long marker = 0; marker<nmarkers; marker++) {
    for (unsigned long ind=0; ind<m_individuals.size(); ind++) {
      if (m_individuals[ind].ChooseRandomHaplotypesFor(lname, marker)) {
        //The chosen one is different from last time.
        ReassignDLsFor(lname, marker, ind);
      }
    }
  }
}

void RecTree::SimulateDataIfNeeded()
{
  Tree::SimulateDataIfNeeded();
  //LS DEBUG: Current version uses same data for all moving loci.  If we want
  // to simulate data individually, uncomment the following lines and replace
  // all the subsequent loops over lnums to locus.whatever calls.
  //
  //  for (unsigned long mloc=0; mloc<m_pMovingLocusVec->size(); mloc++) {
  //  Locus& locus = (*m_pMovingLocusVec)[mloc];
  if ((*m_pMovingLocusVec).size() > 0) {
    Locus& locus = (*m_pMovingLocusVec)[0];
    if (locus.GetShouldSimulate()) {
      for (size_t lnum=0; lnum<(*m_pMovingLocusVec).size(); lnum++) {
        (*m_pMovingLocusVec)[lnum].ClearVariability();
        (*m_pMovingLocusVec)[lnum].SetNewMappositionIfMoving();
      }
      //Check to see if this locus's site was already simulated
      Locus* simlocus = NULL; //to prevent compiler warnings
      for (unsigned long loc=0; loc<m_pLocusVec->size(); loc++) {
        Locus& oldlocus = (*m_pLocusVec)[loc];
        if (oldlocus.GetShouldSimulate()) {
          if (oldlocus.SiteInLocus(locus.GetRegionalMapposition())) {
            simlocus = &oldlocus;
          }
        }
      }
      if (simlocus != NULL) {
        //If there was at least one site in the original that overlapped with
        // the range on the moving locus, we take that chances as the chance
        // that we hit the original at all, but we want to choose a variable
        // site from that range.  This introduces an ascertainment bias
        // that I'm not exactly sure how to compensate for, or if it needs
        // to be compensated for.  But anyway.
        long newsite=simlocus->ChooseVariableSiteFrom(locus.GetAllowedRange());
        if (newsite != FLAGLONG) {
          for (size_t lnum=0; lnum<(*m_pMovingLocusVec).size(); lnum++) {
            (*m_pMovingLocusVec)[lnum].SetRegionalMapposition(newsite);
            (*m_pMovingLocusVec)[lnum].SetTrueSite(newsite);
            //LS TEST
            //simlocus->PrintOnesAndZeroesForVariableSites();
            (*m_pMovingLocusVec)[lnum].SetVariableRange(simlocus->CalculateVariableRange());
          }
          //LS DEBUG:  Not really for release.
          //simlocus->CalculateDisEqFor(newsite);
          //If there were no variable sites in the original, we keep the
          // originally chosen site.  Our trait isn't going to be variable,
          // but them's the breaks.
        }
        for (size_t lnum=0; lnum<(*m_pMovingLocusVec).size(); lnum++) {
          (*m_pMovingLocusVec)[lnum].CopyDataFrom(*simlocus, *this);
        }
        if (locus.IsNighInvariant()) {
          registry.GetRunReport().ReportNormal("All simulated data was nigh invariant for the source segment, giving us nigh invariant data for the simulated segment " + locus.GetName() + " as well.");
        }
        //LS DEBUG SIM:  This would delete information again, for simulations
        // where the data exists, but you never sequenced it.
        // simlocus->RandomizeHalf(*this);
      }
      else {
        locus.SimulateData(*this, m_totalsites);
        //Now copy this data into any other moving loci.
        long newsite = locus.GetRegionalMapposition();
        for (size_t lnum=1; lnum<(*m_pMovingLocusVec).size(); lnum++) {
          (*m_pMovingLocusVec)[lnum].SetRegionalMapposition(newsite);
          (*m_pMovingLocusVec)[lnum].SetTrueSite(newsite);
          (*m_pMovingLocusVec)[lnum].CopyDataFrom(locus, *this);
        }
      }
      for (size_t lnum=0; lnum<(*m_pMovingLocusVec).size(); lnum++) {
        (*m_pMovingLocusVec)[lnum].MakePhenotypesFor(m_individuals);
      }
    }
  }
  //Redo the aliases, in case we randomized the data in the original locus.
  SetupAliases(*m_pLocusVec);
}

long RecTree::NumberOfRecombinations()
{
  return (m_timelist.HowMany(btypeRec)/2);
}

//____________________________________________________________________

void RecTree::SetActiveLinksFrom(const BranchBuffer& brbuffer)
{
activeLinks = 0;

vector<Branch_ptr> branches(brbuffer.ExtractConstBranches());
vector<Branch_ptr>::const_iterator branch;
for(branch = branches.begin(); branch != branches.end(); ++branch)
   activeLinks += (*branch)->m_range.ActiveLinks();

}

//____________________________________________________________________

void RecTree::PrintTipData(long mloc, long marker)
{
  cout << endl << "Tip data:" << endl;
  for (Branchconstiter tip = m_timelist.FirstTip();
       tip!= m_timelist.FirstBody(); tip++) {
    cout << " " << (*m_pMovingLocusVec)[mloc].GetDataModel()->CellToData((*tip)->GetDLCell(mloc, marker, true), marker);
  }
  cout << endl;
}
