//$Id: summary.h,v 1.26 2005/04/26 00:01:34 lpsmith Exp $

#ifndef SUMMARY_H
#define SUMMARY_H

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include <iostream>
#include <string>
#include "vectorx.h"
#include <list>
#include "constants.h"
#include "intervaldata.h"   // for IntervalData member
#include "registry.h"       // for use of GetDataPack() and
                            // GetForceSummary() to set npop & nparams

/*
   This is a Force-polymorphic class which works with the IntervalData
   class (non-polymorphic) to provide needed summary information about
   a tree.  It can provide either Short or Long forms of its summary,
   depending on the presence or absence of perturbing forces like Growth
   (if Growth is in effect Long forms are necessary).

   Written by Mary Kuhner
*/

class Interval;

class Summary
{
private:
  Summary(const Summary&); // not defined
  Summary& operator=(const Summary&); // not defined

protected:

  IntervalData& intervalData;
  DoubleVec1d shortpoint;    // pre-computed ShortPoint results
                             // dimensioned by # of parameters
  DoubleVec1d shortwait;     // pre-computed ShortWait results
                             // dimensioned by # of parameters
  long nparams;
  bool shortness;
  IntervalData fakeIntervals;

  // pointers into the linked list in IntervalData, or possibly
  // into fakeIntervals instead
  Interval* front;   // front interval of our type
  Interval* back;    // most recently added interval of our type

  virtual void ComputeShortPoint() = 0;
  virtual void ComputeShortWait() = 0;

public:

  Summary(IntervalData& interval, long npars, bool shortform) 
    : intervalData(interval), nparams(npars), shortness(shortform),
      front(NULL), back(NULL) {};
  virtual ~Summary() {};

  virtual Summary* Clone(IntervalData& interval) const = 0;

  const DoubleVec1d& GetShortPoint() const; // nCoalescences, etc.
  const DoubleVec1d& GetShortWait() const;  // k(k-1)t, etc.

  Interval const * GetLongPoint() const;
  const std::list<Interval>& GetLongWait() const;

  void AddInterval(double time, const LongVec2d& pk, 
		   const LongVec1d& xk, long s, xpart_t ostat, xpart_t nstat,
		   long recsite, force_type type);

  virtual void AdjustSummary(const DoubleVec1d& totals, long region) = 0;

  bool Compress();

  virtual string GetType() const = 0;
  const bool GetShortness();

  // Functions for reading in data from a summary file:
  void AddShortPoint(string);
  void AddShortWait(string);
  void SetShortness( bool a ) { shortness = a; };
};

//_____________________________________________________________

class CoalSummary : public Summary
{
private:

  virtual void ComputeShortPoint();  // nCoalescences, etc.
  virtual void ComputeShortWait();   // k(k-1)t, etc.

public:

  CoalSummary(IntervalData& interval, bool shortform) 
    : Summary(interval, registry.GetForceSummary().GetNParameters(force_COAL),
              shortform) {};
  virtual ~CoalSummary() {};
  virtual Summary* Clone(IntervalData& interval) const;
  virtual void AdjustSummary(const DoubleVec1d& totals, long region);
  virtual string GetType() const { return lamarcstrings::COAL; };

};

//_____________________________________________________________

class MigSummary : public Summary
{
private:

  virtual void ComputeShortPoint(); 
  virtual void ComputeShortWait(); 
  xpart_t m_npop;
  xpart_t m_migpartindex;

public:

  MigSummary(IntervalData& interval, bool shortform) 
    : Summary(interval, registry.GetForceSummary().GetNParameters(force_MIG),
      shortform),
      m_npop(registry.GetDataPack().GetNPartitionsByForceType(force_MIG)),
      m_migpartindex(registry.GetForceSummary().GetPartIndex(force_MIG)) {};
  virtual ~MigSummary() {};
  virtual Summary* Clone(IntervalData& interval) const;
  virtual void AdjustSummary(const DoubleVec1d& totals, long region);
  virtual string GetType() const { return lamarcstrings::MIG; };

};

//_____________________________________________________________

class DiseaseSummary : public Summary
{
private:
  virtual void ComputeShortPoint();
  virtual void ComputeShortWait();
  xpart_t m_nstati;
  xpart_t m_dispartindex;

public:
  DiseaseSummary(IntervalData& interval, bool shortform) :
    Summary(interval, registry.GetForceSummary().GetNParameters(force_DISEASE),
    shortform),
    m_nstati(registry.GetDataPack().GetNPartitionsByForceType(force_DISEASE)),
    m_dispartindex(registry.GetForceSummary().GetPartIndex(force_DISEASE)) {};
  virtual ~DiseaseSummary() {};
  virtual Summary* Clone(IntervalData& interval) const;
  virtual void AdjustSummary(const DoubleVec1d& totals, long region);
  virtual string GetType() const { return lamarcstrings::DISEASE; };

};

//_____________________________________________________________

class RecSummary : public Summary
{

private:

  virtual void ComputeShortPoint(); 
  virtual void ComputeShortWait(); 

public:
  RecSummary(IntervalData& interval, bool shortform) 
    : Summary(interval, registry.GetForceSummary().GetNParameters(force_REC),
              shortform) {};
  virtual ~RecSummary() {};
  virtual Summary* Clone(IntervalData& interval) const;
  virtual void AdjustSummary(const DoubleVec1d& totals, long region);
  virtual string GetType() const { return lamarcstrings::REC; };

};

#endif /* SUMMARY_H */
